# Copyright (c) 2014 Andrew Kerr.  All rights reserved.
# Copyright (c) 2015 Tom Barron.  All rights reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the NetApp NFS storage driver."""
import copy
import os
import threading
import time
from unittest import mock

import ddt
from os_brick.remotefs import remotefs as remotefs_brick
from oslo_concurrency import processutils
from oslo_utils import units

from cinder import context
from cinder import exception
from cinder.objects import fields
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.netapp.dataontap import fakes as fake
from cinder import utils
from cinder.volume.drivers.netapp.dataontap.client import api as netapp_api
from cinder.volume.drivers.netapp.dataontap import nfs_base
from cinder.volume.drivers.netapp.dataontap.utils import loopingcalls
from cinder.volume.drivers.netapp import utils as na_utils
from cinder.volume.drivers import nfs
from cinder.volume.drivers import remotefs
from cinder.volume import volume_utils


@ddt.ddt
class NetAppNfsDriverTestCase(test.TestCase):
    def setUp(self):
        super(NetAppNfsDriverTestCase, self).setUp()
        configuration = mock.Mock()
        configuration.reserved_percentage = 0
        configuration.nfs_mount_point_base = '/mnt/test'
        configuration.reserved_percentage = 0
        configuration.max_over_subscription_ratio = 1.1
        self.fake_nfs_export_1 = fake.NFS_EXPORT_1
        self.fake_nfs_export_2 = fake.NFS_EXPORT_2
        self.fake_mount_point = fake.MOUNT_POINT
        self.ctxt = context.RequestContext('fake', 'fake', auth_token=True)

        kwargs = {
            'configuration': configuration,
            'host': 'openstack@netappnfs',
        }

        with mock.patch.object(utils, 'get_root_helper',
                               return_value=mock.Mock()):
            with mock.patch.object(remotefs_brick, 'RemoteFsClient',
                                   return_value=mock.Mock()):
                self.driver = nfs_base.NetAppNfsDriver(**kwargs)
                self.driver.db = mock.Mock()

        self.driver.zapi_client = mock.Mock()
        self.zapi_client = self.driver.zapi_client

    @mock.patch.object(nfs.NfsDriver, 'do_setup')
    @mock.patch.object(na_utils, 'check_flags')
    def test_do_setup(self, mock_check_flags, mock_super_do_setup):
        self.driver.do_setup(mock.Mock())

        self.assertTrue(mock_check_flags.called)
        self.assertTrue(mock_super_do_setup.called)

    def test_get_share_capacity_info(self):
        mock_get_capacity = self.mock_object(self.driver, '_get_capacity_info')
        mock_get_capacity.return_value = fake.CAPACITY_VALUES
        expected_total_capacity_gb = na_utils.round_down(
            fake.TOTAL_BYTES / units.Gi, '0.01')
        expected_free_capacity_gb = (na_utils.round_down(
            fake.AVAILABLE_BYTES / units.Gi, '0.01'))
        expected_reserved_percentage = round(
            self.driver.configuration.reserved_percentage)

        result = self.driver._get_share_capacity_info(fake.NFS_SHARE)

        self.assertEqual(expected_total_capacity_gb,
                         result['total_capacity_gb'])
        self.assertEqual(expected_free_capacity_gb,
                         result['free_capacity_gb'])
        self.assertEqual(expected_reserved_percentage,
                         round(result['reserved_percentage']))

    def test_get_capacity_info_ipv4_share(self):
        expected = fake.CAPACITY_VALUES
        get_capacity = self.driver.zapi_client.get_flexvol_capacity
        get_capacity.return_value = fake.CAPACITIES

        result = self.driver._get_capacity_info(fake.NFS_SHARE_IPV4)

        self.assertEqual(expected, result)
        get_capacity.assert_has_calls([
            mock.call(flexvol_path=fake.EXPORT_PATH)])

    def test_get_capacity_info_ipv6_share(self):
        expected = fake.CAPACITY_VALUES
        get_capacity = self.driver.zapi_client.get_flexvol_capacity
        get_capacity.return_value = fake.CAPACITIES

        result = self.driver._get_capacity_info(fake.NFS_SHARE_IPV6)

        self.assertEqual(expected, result)
        get_capacity.assert_has_calls([
            mock.call(flexvol_path=fake.EXPORT_PATH)])

    def test_get_pool(self):
        pool = self.driver.get_pool({'provider_location': 'fake-share'})

        self.assertEqual('fake-share', pool)

    @ddt.data(None,
              {'replication_status': fields.ReplicationStatus.ENABLED})
    def test_create_volume(self, model_update):
        self.mock_object(self.driver, '_ensure_shares_mounted')
        self.mock_object(self.driver, '_ensure_flexgroup_not_in_cg')
        self.mock_object(na_utils, 'get_volume_extra_specs')
        self.mock_object(self.driver, '_do_create_volume')
        self.mock_object(self.driver, '_do_qos_for_volume')
        self.mock_object(self.driver, '_get_volume_model_update',
                         return_value=model_update)
        expected = {'provider_location': fake.NFS_SHARE}
        if model_update:
            expected.update(model_update)

        actual = self.driver.create_volume(fake.NFS_VOLUME)

        self.assertEqual(expected, actual)

    def test_create_volume_no_pool(self):
        volume = copy.deepcopy(fake.NFS_VOLUME)
        volume['host'] = '%s@%s' % (fake.HOST_NAME, fake.BACKEND_NAME)
        self.mock_object(self.driver, '_ensure_shares_mounted')
        self.mock_object(self.driver, '_ensure_flexgroup_not_in_cg')

        self.assertRaises(exception.InvalidHost,
                          self.driver.create_volume,
                          volume)

    def test_create_volume_exception(self):
        self.mock_object(self.driver, '_ensure_shares_mounted')
        self.mock_object(self.driver, '_ensure_flexgroup_not_in_cg')
        self.mock_object(na_utils, 'get_volume_extra_specs')
        mock_create = self.mock_object(self.driver, '_do_create_volume')
        mock_create.side_effect = Exception

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume,
                          fake.NFS_VOLUME)

    @ddt.data(None, {'key': 'value'})
    def test_clone_source_to_destination_volume(self, model_update):
        self.mock_object(self.driver, '_get_volume_location',
                         return_value=fake.POOL_NAME)
        self.mock_object(na_utils, 'get_volume_extra_specs',
                         return_value=fake.EXTRA_SPECS)
        self.mock_object(
            self.driver,
            '_clone_with_extension_check')
        self.mock_object(self.driver, '_do_qos_for_volume')
        self.mock_object(self.driver, '_get_volume_model_update',
                         return_value=model_update)
        expected = {'provider_location': fake.POOL_NAME}
        if model_update:
            expected.update(model_update)

        result = self.driver._clone_source_to_destination_volume(
            fake.CLONE_SOURCE, fake.CLONE_DESTINATION)

        self.assertEqual(expected, result)

    def test_clone_source_to_destination_volume_with_do_qos_exception(self):
        self.mock_object(self.driver, '_get_volume_location',
                         return_value=fake.POOL_NAME)
        self.mock_object(na_utils, 'get_volume_extra_specs',
                         return_value=fake.EXTRA_SPECS)
        self.mock_object(
            self.driver,
            '_clone_with_extension_check')
        self.mock_object(self.driver, '_do_qos_for_volume',
                         side_effect=Exception)

        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.driver._clone_source_to_destination_volume,
            fake.CLONE_SOURCE,
            fake.CLONE_DESTINATION)

    def test_clone_with_extension_check_equal_sizes(self):
        clone_source = copy.deepcopy(fake.CLONE_SOURCE)
        clone_source['size'] = fake.VOLUME['size']
        self.mock_object(self.driver, '_clone_backing_file_for_volume')
        self.mock_object(self.driver, 'local_path')
        mock_discover = self.mock_object(self.driver,
                                         '_discover_file_till_timeout')
        mock_discover.return_value = True
        self.mock_object(self.driver, '_set_rw_permissions')
        mock_extend_volume = self.mock_object(self.driver, 'extend_volume')

        self.driver._clone_with_extension_check(clone_source, fake.NFS_VOLUME)

        self.assertEqual(0, mock_extend_volume.call_count)

    def test_clone_with_extension_check_unequal_sizes(self):
        clone_source = copy.deepcopy(fake.CLONE_SOURCE)
        clone_source['size'] = fake.VOLUME['size'] + 1
        self.mock_object(self.driver, '_clone_backing_file_for_volume')
        self.mock_object(self.driver, 'local_path')
        mock_discover = self.mock_object(self.driver,
                                         '_discover_file_till_timeout')
        mock_discover.return_value = True
        self.mock_object(self.driver, '_set_rw_permissions')
        mock_extend_volume = self.mock_object(self.driver, 'extend_volume')

        self.driver._clone_with_extension_check(clone_source, fake.NFS_VOLUME)

        self.assertEqual(1, mock_extend_volume.call_count)

    def test_clone_with_extension_check_extend_exception(self):
        clone_source = copy.deepcopy(fake.CLONE_SOURCE)
        clone_source['size'] = fake.VOLUME['size'] + 1
        self.mock_object(self.driver, '_clone_backing_file_for_volume')
        self.mock_object(self.driver, 'local_path')
        mock_discover = self.mock_object(self.driver,
                                         '_discover_file_till_timeout')
        mock_discover.return_value = True
        self.mock_object(self.driver, '_set_rw_permissions')
        mock_extend_volume = self.mock_object(self.driver, 'extend_volume')
        mock_extend_volume.side_effect = Exception
        mock_cleanup = self.mock_object(self.driver,
                                        '_cleanup_volume_on_failure')

        self.assertRaises(exception.CinderException,
                          self.driver._clone_with_extension_check,
                          clone_source,
                          fake.NFS_VOLUME)

        self.assertEqual(1, mock_cleanup.call_count)

    def test_clone_with_extension_check_no_discovery(self):
        self.mock_object(self.driver, '_clone_backing_file_for_volume')
        self.mock_object(self.driver, 'local_path')
        self.mock_object(self.driver, '_set_rw_permissions')
        mock_discover = self.mock_object(self.driver,
                                         '_discover_file_till_timeout')
        mock_discover.return_value = False

        self.assertRaises(exception.CinderException,
                          self.driver._clone_with_extension_check,
                          fake.CLONE_SOURCE,
                          fake.NFS_VOLUME)

    @ddt.data(True, False)
    def test_create_volume_from_snapshot(self, is_flexgroup):
        provider_location = fake.POOL_NAME
        volume = fake.VOLUME
        expected_source = {
            'name': fake.SNAPSHOT_NAME,
            'size': fake.SIZE,
            'id': fake.VOLUME_ID,
        }
        mock_clone_call = self.mock_object(
            self.driver, '_clone_source_to_destination_volume',
            return_value=provider_location)
        self.mock_object(self.driver, '_ensure_flexgroup_not_in_cg')
        self.mock_object(self.driver, '_is_flexgroup',
                         return_value=is_flexgroup)
        self.mock_object(self.driver, '_is_flexgroup_clone_file_supported',
                         return_value=not is_flexgroup)
        mock_super_create = self.mock_object(
            nfs.NfsDriver, 'create_volume_from_snapshot',
            return_value=provider_location)
        mock_do_qos = self.mock_object(
            self.driver, '_do_qos_for_file_flexgroup',
            return_value=provider_location)

        retval = self.driver.create_volume_from_snapshot(volume, fake.SNAPSHOT)

        self.assertEqual(provider_location, retval)
        if is_flexgroup:
            mock_clone_call.assert_not_called()
            mock_super_create.assert_called_once_with(volume, fake.SNAPSHOT)
            mock_do_qos.assert_called_once_with(volume, provider_location)
        else:
            mock_clone_call.assert_called_once_with(expected_source, volume)
            mock_do_qos.assert_not_called()
            mock_super_create.not_called()

    @ddt.data(True, False)
    def test_create_cloned_volume(self, is_flexgroup):
        provider_location = fake.POOL_NAME
        volume = fake.VOLUME
        src_vref = fake.CLONE_SOURCE
        mock_clone_call = self.mock_object(
            self.driver, '_clone_source_to_destination_volume',
            return_value=provider_location)
        self.mock_object(self.driver, '_ensure_flexgroup_not_in_cg')
        self.mock_object(self.driver, '_is_flexgroup',
                         return_value=is_flexgroup)
        self.mock_object(self.driver, '_is_flexgroup_clone_file_supported',
                         return_value=not is_flexgroup)
        mock_super_create = self.mock_object(
            nfs.NfsDriver, 'create_cloned_volume',
            return_value=provider_location)
        mock_do_qos = self.mock_object(
            self.driver, '_do_qos_for_file_flexgroup',
            return_value=provider_location)

        result = self.driver.create_cloned_volume(volume, src_vref)

        self.assertEqual(provider_location, result)
        if is_flexgroup:
            mock_clone_call.assert_not_called()
            mock_super_create.assert_called_once_with(volume, src_vref)
            mock_do_qos.assert_called_once_with(volume, provider_location)
        else:
            mock_clone_call.assert_called_once_with(src_vref, volume)
            mock_do_qos.assert_not_called()
            mock_super_create.not_called()

    def test_do_qos_for_file_flexgroup(self):
        volume = {'provider_location': 'fake'}
        extra_specs = 'fake_extra'
        model = {'provider_location': 'fake'}
        vol_model = {'replication': 'fake'}
        expected_model = {
            'replication': vol_model['replication'],
            'provider_location': model['provider_location'],
        }
        self.mock_object(self.driver, '_do_qos_for_volume')
        self.mock_object(self.driver, '_get_volume_model_update',
                         return_value=vol_model)
        mock_extra = self.mock_object(na_utils, 'get_volume_extra_specs',
                                      return_value=extra_specs)

        model_updated = self.driver._do_qos_for_file_flexgroup(volume, model)

        self.assertEqual(model_updated, expected_model)
        mock_extra.assert_called_once_with(volume)

    def test_do_qos_for_file_flexgroup_error(self):
        self.mock_object(na_utils, 'get_volume_extra_specs',
                         side_effect=exception.NotFound)

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver._do_qos_for_file_flexgroup,
                          fake.VOLUME, 'fake_model')

    def test_do_qos_for_volume(self):
        self.assertRaises(NotImplementedError,
                          self.driver._do_qos_for_volume,
                          fake.NFS_VOLUME,
                          fake.EXTRA_SPECS)

    @ddt.data(True, False)
    def test_create_snapshot(self, is_flexgroup):
        self.mock_object(self.driver, '_is_flexgroup',
                         return_value=is_flexgroup)
        self.mock_object(self.driver, '_is_flexgroup_clone_file_supported',
                         return_value=not is_flexgroup)
        mock_clone_backing_file_for_volume = self.mock_object(
            self.driver, '_clone_backing_file_for_volume')
        mock_snap_flexgroup = self.mock_object(
            self.driver, '_create_snapshot_for_flexgroup')

        self.driver.create_snapshot(fake.SNAPSHOT)

        if is_flexgroup:
            mock_snap_flexgroup.assert_called_once_with(fake.SNAPSHOT)
            mock_clone_backing_file_for_volume.assert_not_called()
        else:
            mock_snap_flexgroup.assert_not_called()
            mock_clone_backing_file_for_volume.assert_called_once_with(
                fake.SNAPSHOT['volume_name'], fake.SNAPSHOT['name'],
                fake.SNAPSHOT['volume_id'], is_snapshot=True)

    def test_create_snapshot_for_flexgroup(self):
        source_vol = {
            'id': fake.SNAPSHOT['volume_id'],
            'name': fake.SNAPSHOT['volume_name'],
            'volume_type_id': fake.SNAPSHOT['volume_type_id'],
        }
        snap_vol = {
            'name': '%s.%s' % (fake.SNAPSHOT['volume_name'],
                               fake.SNAPSHOT['id']),
            'host': fake.HOST_NAME,
        }
        mock_super_snapshot = self.mock_object(nfs.NfsDriver,
                                               'create_snapshot')
        mock_extra_specs = self.mock_object(na_utils,
                                            'get_volume_extra_specs')
        mock_extra_specs.return_value = fake.EXTRA_SPECS
        mock_get_info = self.mock_object(na_utils,
                                         'get_valid_qos_policy_group_info')
        mock_get_info.return_value = fake.QOS_POLICY_GROUP_INFO
        mock_get_host = self.mock_object(self.driver,
                                         '_get_volume_host')
        mock_get_host.return_value = fake.HOST_NAME
        mock_set_policy = self.mock_object(self.driver,
                                           '_set_qos_policy_group_on_volume')

        self.driver._create_snapshot_for_flexgroup(fake.SNAPSHOT)

        mock_super_snapshot.assert_has_calls([
            mock.call(fake.SNAPSHOT)])
        mock_get_host.assert_has_calls([
            mock.call(source_vol['id'])])
        mock_extra_specs.assert_has_calls([
            mock.call(source_vol)])
        mock_get_info.assert_has_calls([
            mock.call(source_vol, fake.EXTRA_SPECS)])
        mock_set_policy.assert_has_calls([
            mock.call(snap_vol, fake.QOS_POLICY_GROUP_INFO, False)])

    def test_create_snapshot_for_flexgroup_error(self):
        self.mock_object(nfs.NfsDriver, 'create_snapshot',
                         side_effect=exception.NotFound)

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver._create_snapshot_for_flexgroup,
                          fake.SNAPSHOT)

    def test_set_qos_policy_group_on_volume(self):
        self.assertRaises(NotImplementedError,
                          self.driver._set_qos_policy_group_on_volume,
                          fake.NFS_VOLUME,
                          fake.QOS_POLICY_GROUP_INFO,
                          False)

    @ddt.data(True, False)
    def test_delete_snapshot(self, is_flexgroup):
        updates = {
            'name': fake.SNAPSHOT_NAME,
            'volume_size': fake.SIZE,
            'volume_id': fake.VOLUME_ID,
            'volume_name': fake.VOLUME_NAME,
            'busy': False,
        }
        snapshot = fake_snapshot.fake_snapshot_obj(self.ctxt, **updates)
        self.mock_object(self.driver, '_delete_file')
        self.mock_object(self.driver, '_is_flexgroup',
                         return_value=is_flexgroup)
        self.mock_object(self.driver, '_is_flexgroup_clone_file_supported',
                         return_value=not is_flexgroup)
        mock_super_delete = self.mock_object(nfs.NfsDriver,
                                             'delete_snapshot')

        self.driver.delete_snapshot(snapshot)

        if is_flexgroup:
            mock_super_delete.assert_called_once_with(snapshot)
            self.driver._delete_file.assert_not_called()
        else:
            mock_super_delete.assert_not_called()
            self.driver._delete_file.assert_called_once_with(
                snapshot.volume_id, snapshot.name)

    @ddt.data(fake.NFS_SHARE, fake.NFS_SHARE_IPV6)
    def test__get_volume_location(self, provider):
        volume_id = fake.VOLUME_ID

        self.mock_object(self.driver, '_get_provider_location',
                         return_value=provider)

        retval = self.driver._get_volume_location(volume_id)

        self.assertEqual(provider, retval)

    def test__clone_backing_file_for_volume(self):
        self.assertRaises(NotImplementedError,
                          self.driver._clone_backing_file_for_volume,
                          fake.VOLUME_NAME, fake.CLONE_SOURCE_NAME,
                          fake.VOLUME_ID, share=None)

    def test__is_flexgroup(self):
        self.assertRaises(NotImplementedError,
                          self.driver._is_flexgroup)

    def test__get_provider_location(self):
        updates = {'provider_location': fake.PROVIDER_LOCATION}
        volume = fake_volume.fake_volume_obj(self.ctxt, **updates)
        self.mock_object(self.driver.db, 'volume_get', return_value=volume)

        retval = self.driver._get_provider_location(fake.VOLUME_ID)

        self.assertEqual(fake.PROVIDER_LOCATION, retval)

    def test__get_volume_host(self):
        updates = {'host': fake.HOST_NAME}
        volume = fake_volume.fake_volume_obj(self.ctxt, **updates)
        self.mock_object(self.driver.db, 'volume_get', return_value=volume)

        retval = self.driver._get_volume_host(fake.VOLUME_ID)

        self.assertEqual(fake.HOST_NAME, retval)

    @ddt.data(None, processutils.ProcessExecutionError)
    def test__volume_not_present(self, side_effect):
        self.mock_object(self.driver, '_get_volume_path')
        self.mock_object(self.driver, '_try_execute', side_effect=side_effect)

        retval = self.driver._volume_not_present(
            fake.MOUNT_PATH, fake.VOLUME_NAME)

        self.assertEqual(side_effect is not None, retval)

    @mock.patch.object(time, 'sleep')
    def test__try_execute_exception(self, patched_sleep):
        self.mock_object(self.driver, '_execute',
                         side_effect=processutils.ProcessExecutionError)
        mock_exception_log = self.mock_object(nfs_base.LOG, 'exception')
        self.driver.configuration.num_shell_tries = 3

        self.assertRaises(processutils.ProcessExecutionError,
                          self.driver._try_execute,
                          'fake-command', attr1='val1', attr2='val2')
        self.assertEqual(2, mock_exception_log.call_count)
        self.driver._execute.assert_has_calls([
            mock.call('fake-command', attr1='val1', attr2='val2'),
            mock.call('fake-command', attr1='val1', attr2='val2'),
            mock.call('fake-command', attr1='val1', attr2='val2')])
        self.assertEqual(2, time.sleep.call_count)
        patched_sleep.assert_has_calls([mock.call(1), mock.call(4)])

    def test__update_volume_stats(self):
        self.assertRaises(NotImplementedError,
                          self.driver._update_volume_stats)

    def test_copy_image_to_volume_base_exception(self):
        mock_info_log = self.mock_object(nfs_base.LOG, 'info')
        self.mock_object(self.driver, '_ensure_flexgroup_not_in_cg')
        self.mock_object(remotefs.RemoteFSDriver, 'copy_image_to_volume',
                         side_effect=exception.NfsException)

        self.assertRaises(exception.NfsException,
                          self.driver.copy_image_to_volume,
                          'fake_context', fake.NFS_VOLUME,
                          'fake_img_service', fake.IMAGE_FILE_ID)
        mock_info_log.assert_not_called()

    def test_copy_image_to_volume(self):
        mock_log = self.mock_object(nfs_base, 'LOG')
        self.mock_object(self.driver, '_is_flexgroup',
                         return_value=False)
        self.mock_object(self.driver, '_is_flexgroup_clone_file_supported',
                         return_value=True)
        self.mock_object(self.driver, '_ensure_flexgroup_not_in_cg')
        mock_copy_image = self.mock_object(
            remotefs.RemoteFSDriver, 'copy_image_to_volume')
        mock_register_image = self.mock_object(
            self.driver, '_register_image_in_cache')

        self.driver.copy_image_to_volume('fake_context',
                                         fake.NFS_VOLUME,
                                         'fake_img_service',
                                         fake.IMAGE_FILE_ID)

        mock_copy_image.assert_called_once_with(
            'fake_context', fake.NFS_VOLUME, 'fake_img_service',
            fake.IMAGE_FILE_ID)
        self.assertEqual(1, mock_log.info.call_count)
        mock_register_image.assert_called_once_with(
            fake.NFS_VOLUME, fake.IMAGE_FILE_ID)

    @ddt.data(None, Exception)
    def test__register_image_in_cache(self, exc):
        mock_log = self.mock_object(nfs_base, 'LOG')
        self.mock_object(self.driver, '_do_clone_rel_img_cache',
                         side_effect=exc)

        retval = self.driver._register_image_in_cache(
            fake.NFS_VOLUME, fake.IMAGE_FILE_ID)

        self.assertIsNone(retval)
        self.assertEqual(exc is not None, mock_log.warning.called)
        self.assertEqual(1, mock_log.info.call_count)

    @ddt.data(True, False)
    def test_do_clone_rel_img_cache(self, path_exists):
        self.mock_object(nfs_base.LOG, 'info')
        self.mock_object(utils, 'synchronized', return_value=lambda f: f)
        self.mock_object(self.driver, '_get_mount_point_for_share',
                         return_value='dir')
        self.mock_object(os.path, 'exists', return_value=path_exists)
        self.mock_object(self.driver, '_clone_backing_file_for_volume')
        self.mock_object(os, 'utime')

        retval = self.driver._do_clone_rel_img_cache(
            fake.CLONE_SOURCE_NAME, fake.CLONE_DESTINATION_NAME,
            fake.NFS_SHARE, 'fake_cache_file')

        self.assertIsNone(retval)
        self.assertTrue(self.driver._get_mount_point_for_share.called)
        if not path_exists:
            self.driver._clone_backing_file_for_volume.assert_called_once_with(
                fake.CLONE_SOURCE_NAME, fake.CLONE_DESTINATION_NAME,
                share=fake.NFS_SHARE, volume_id=None)
            os.utime.assert_called_once_with(
                'dir/' + fake.CLONE_SOURCE_NAME, None)
        else:
            self.driver._clone_backing_file_for_volume.assert_not_called()
            os.utime.assert_not_called()

        os.path.exists.assert_called_once_with(
            'dir/' + fake.CLONE_DESTINATION_NAME)

    def test__spawn_clean_cache_job_clean_job_setup(self):
        self.driver.cleaning = True
        mock_debug_log = self.mock_object(nfs_base.LOG, 'debug')
        self.mock_object(utils, 'synchronized', return_value=lambda f: f)

        retval = self.driver._spawn_clean_cache_job()

        self.assertIsNone(retval)
        self.assertEqual(1, mock_debug_log.call_count)

    def test__spawn_clean_cache_job_new_clean_job(self):

        class FakeTimer(object):
            def start(self):
                pass

        fake_timer = FakeTimer()
        self.mock_object(utils, 'synchronized', return_value=lambda f: f)
        self.mock_object(fake_timer, 'start')
        self.mock_object(nfs_base.LOG, 'debug')
        self.mock_object(self.driver, '_clean_image_cache')
        self.mock_object(threading, 'Timer', return_value=fake_timer)

        retval = self.driver._spawn_clean_cache_job()

        self.assertIsNone(retval)
        threading.Timer.assert_called_once_with(
            0, self.driver._clean_image_cache)
        fake_timer.start.assert_called_once_with()

    def test_cleanup_volume_on_failure(self):
        path = '%s/%s' % (fake.NFS_SHARE, fake.NFS_VOLUME['name'])
        mock_local_path = self.mock_object(self.driver, 'local_path')
        mock_local_path.return_value = path
        mock_exists_check = self.mock_object(os.path, 'exists')
        mock_exists_check.return_value = True
        mock_delete = self.mock_object(self.driver, '_delete_file_at_path')

        self.driver._cleanup_volume_on_failure(fake.NFS_VOLUME)

        mock_delete.assert_has_calls([mock.call(path)])

    def test_cleanup_volume_on_failure_no_path(self):
        self.mock_object(self.driver, 'local_path')
        mock_exists_check = self.mock_object(os.path, 'exists')
        mock_exists_check.return_value = False
        mock_delete = self.mock_object(self.driver, '_delete_file_at_path')

        self.driver._cleanup_volume_on_failure(fake.NFS_VOLUME)

        self.assertEqual(0, mock_delete.call_count)

    @ddt.data((fake.NFS_SHARE, fake.SHARE_IP),
              (fake.NFS_SHARE_IPV6, fake.IPV6_ADDRESS))
    @ddt.unpack
    def test_get_export_ip_path_volume_id_provided(self, provider_location,
                                                   ip):
        mock_get_host_ip = self.mock_object(self.driver,
                                            '_get_provider_location')
        mock_get_host_ip.return_value = provider_location

        expected = (ip, fake.EXPORT_PATH)

        result = self.driver._get_export_ip_path(fake.VOLUME_ID)

        self.assertEqual(expected, result)

    @ddt.data((fake.NFS_SHARE, fake.SHARE_IP, fake.EXPORT_PATH),
              (fake.NFS_SHARE_IPV6, fake.IPV6_ADDRESS, fake.EXPORT_PATH))
    @ddt.unpack
    def test_get_export_ip_path_share_provided(self, share, ip, path):
        expected = (ip, path)

        result = self.driver._get_export_ip_path(share=share)

        self.assertEqual(expected, result)

    def test_get_export_ip_path_volume_id_and_share_provided(self):
        mock_get_host_ip = self.mock_object(self.driver,
                                            '_get_provider_location')
        mock_get_host_ip.return_value = fake.NFS_SHARE_IPV4

        expected = (fake.IPV4_ADDRESS, fake.EXPORT_PATH)

        result = self.driver._get_export_ip_path(
            fake.VOLUME_ID, fake.NFS_SHARE)

        self.assertEqual(expected, result)

    def test_get_export_ip_path_no_args(self):
        self.assertRaises(exception.InvalidInput,
                          self.driver._get_export_ip_path)

    def test_construct_image_url_loc(self):
        img_loc = fake.FAKE_IMAGE_LOCATION

        locations = self.driver._construct_image_nfs_url(img_loc)

        self.assertIn("nfs://host/path/image-id-0", locations)
        self.assertIn("nfs://host/path/image-id-6", locations)
        self.assertEqual(2, len(locations))

    def test_construct_image_url_direct(self):
        img_loc = ("nfs://host/path/image-id", None)

        locations = self.driver._construct_image_nfs_url(img_loc)

        self.assertIn("nfs://host/path/image-id", locations)

    @ddt.data(None, 'raw', 'qcow2')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    def test_extend_volume(self, file_format, mock_get):

        volume = fake_volume.fake_volume_obj(self.ctxt)
        if file_format:
            volume.admin_metadata = {'format': file_format}
        mock_get.return_value = volume
        new_size = 100
        volume_copy = copy.copy(volume)
        volume_copy['size'] = new_size

        path = '%s/%s' % (fake.NFS_SHARE, fake.NFS_VOLUME['name'])
        self.mock_object(self.driver,
                         'local_path',
                         return_value=path)
        mock_resize_image_file = self.mock_object(self.driver,
                                                  '_resize_image_file')
        mock_get_volume_extra_specs = self.mock_object(
            na_utils, 'get_volume_extra_specs', return_value=fake.EXTRA_SPECS)
        mock_do_qos_for_volume = self.mock_object(self.driver,
                                                  '_do_qos_for_volume')

        self.driver.extend_volume(volume, new_size)

        mock_resize_image_file.assert_called_once_with(path, new_size,
                                                       file_format=file_format)
        mock_get_volume_extra_specs.assert_called_once_with(volume)
        mock_do_qos_for_volume.assert_called_once_with(volume_copy,
                                                       fake.EXTRA_SPECS,
                                                       cleanup=False)

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    def test_extend_volume_resize_error(self, mock_get):

        volume = fake_volume.fake_volume_obj(self.ctxt)
        mock_get.return_value = volume
        new_size = 100
        volume_copy = copy.copy(volume)
        volume_copy['size'] = new_size

        path = '%s/%s' % (fake.NFS_SHARE, fake.NFS_VOLUME['name'])
        self.mock_object(self.driver,
                         'local_path',
                         return_value=path)
        mock_resize_image_file = self.mock_object(
            self.driver, '_resize_image_file',
            side_effect=netapp_api.NaApiError)
        mock_get_volume_extra_specs = self.mock_object(
            na_utils, 'get_volume_extra_specs', return_value=fake.EXTRA_SPECS)
        mock_do_qos_for_volume = self.mock_object(self.driver,
                                                  '_do_qos_for_volume')

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.extend_volume,
                          volume,
                          new_size)

        mock_resize_image_file.assert_called_once_with(path, new_size,
                                                       file_format=None)
        self.assertFalse(mock_get_volume_extra_specs.called)
        self.assertFalse(mock_do_qos_for_volume.called)

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    def test_extend_volume_qos_error(self, mock_get):

        volume = fake_volume.fake_volume_obj(self.ctxt)
        mock_get.return_value = volume
        new_size = 100
        volume_copy = copy.copy(volume)
        volume_copy['size'] = new_size

        path = '%s/%s' % (fake.NFS_SHARE, fake.NFS_VOLUME['name'])
        self.mock_object(self.driver,
                         'local_path',
                         return_value=path)
        mock_resize_image_file = self.mock_object(self.driver,
                                                  '_resize_image_file')
        mock_get_volume_extra_specs = self.mock_object(
            na_utils, 'get_volume_extra_specs',
            return_value=fake.EXTRA_SPECS)
        mock_do_qos_for_volume = self.mock_object(
            self.driver, '_do_qos_for_volume',
            side_effect=netapp_api.NaApiError)

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.extend_volume,
                          volume,
                          new_size)

        mock_resize_image_file.assert_called_once_with(path, new_size,
                                                       file_format=None)
        mock_get_volume_extra_specs.assert_called_once_with(volume)
        mock_do_qos_for_volume.assert_called_once_with(volume_copy,
                                                       fake.EXTRA_SPECS,
                                                       cleanup=False)

    def test_is_share_clone_compatible(self):
        self.assertRaises(NotImplementedError,
                          self.driver._is_share_clone_compatible,
                          fake.NFS_VOLUME,
                          fake.NFS_SHARE)

    def test_get_share_mount_and_vol_from_vol_ref(self):
        self.mock_object(volume_utils, 'resolve_hostname',
                         return_value='10.12.142.11')
        self.mock_object(os.path, 'isfile', return_value=True)
        self.driver._mounted_shares = [self.fake_nfs_export_1]
        vol_path = "%s/%s" % (self.fake_nfs_export_1, 'test_file_name')
        vol_ref = {'source-name': vol_path}
        self.driver._ensure_shares_mounted = mock.Mock()
        self.driver._get_mount_point_for_share = mock.Mock(
            return_value=self.fake_mount_point)

        (share, mount, file_path) = (
            self.driver._get_share_mount_and_vol_from_vol_ref(vol_ref))

        self.assertEqual(self.fake_nfs_export_1, share)
        self.assertEqual(self.fake_mount_point, mount)
        self.assertEqual('test_file_name', file_path)

    def test_get_share_mount_and_vol_from_vol_ref_with_bad_ref(self):
        self.mock_object(volume_utils, 'resolve_hostname',
                         return_value='10.12.142.11')
        self.driver._mounted_shares = [self.fake_nfs_export_1]
        vol_ref = {'source-id': '1234546'}

        self.driver._ensure_shares_mounted = mock.Mock()
        self.driver._get_mount_point_for_share = mock.Mock(
            return_value=self.fake_mount_point)

        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver._get_share_mount_and_vol_from_vol_ref,
                          vol_ref)

    def test_get_share_mount_and_vol_from_vol_ref_where_not_found(self):
        self.mock_object(volume_utils, 'resolve_hostname',
                         return_value='10.12.142.11')
        self.driver._mounted_shares = [self.fake_nfs_export_1]
        vol_path = "%s/%s" % (self.fake_nfs_export_2, 'test_file_name')
        vol_ref = {'source-name': vol_path}

        self.driver._ensure_shares_mounted = mock.Mock()
        self.driver._get_mount_point_for_share = mock.Mock(
            return_value=self.fake_mount_point)

        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver._get_share_mount_and_vol_from_vol_ref,
                          vol_ref)

    def test_get_share_mount_and_vol_from_vol_ref_where_is_dir(self):
        self.mock_object(volume_utils, 'resolve_hostname',
                         return_value='10.12.142.11')
        self.driver._mounted_shares = [self.fake_nfs_export_1]
        vol_ref = {'source-name': self.fake_nfs_export_2}

        self.driver._ensure_shares_mounted = mock.Mock()
        self.driver._get_mount_point_for_share = mock.Mock(
            return_value=self.fake_mount_point)

        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver._get_share_mount_and_vol_from_vol_ref,
                          vol_ref)

    @ddt.data(None,
              {'replication_status': fields.ReplicationStatus.ENABLED})
    def test_manage_existing(self, model_update):
        self.mock_object(utils, 'get_file_size',
                         return_value=1074253824)
        self.driver._mounted_shares = [self.fake_nfs_export_1]
        test_file = 'test_file_name'
        volume = fake.FAKE_MANAGE_VOLUME
        vol_path = "%s/%s" % (self.fake_nfs_export_1, test_file)
        vol_ref = {'source-name': vol_path}
        self.driver._check_volume_type = mock.Mock()
        self.mock_object(self.driver, '_execute')
        self.driver._ensure_shares_mounted = mock.Mock()
        self.driver._get_mount_point_for_share = mock.Mock(
            return_value=self.fake_mount_point)
        self.driver._get_share_mount_and_vol_from_vol_ref = mock.Mock(
            return_value=(self.fake_nfs_export_1, self.fake_mount_point,
                          test_file))
        mock_get_specs = self.mock_object(na_utils, 'get_volume_extra_specs')
        mock_get_specs.return_value = {}
        self.mock_object(self.driver, '_do_qos_for_volume')
        self.mock_object(self.driver, '_get_volume_model_update',
                         return_value=model_update)

        actual_model_update = self.driver.manage_existing(volume, vol_ref)

        self.assertEqual(
            self.fake_nfs_export_1, actual_model_update['provider_location'])
        if model_update:
            self.assertEqual(model_update['replication_status'],
                             actual_model_update['replication_status'])
        else:
            self.assertNotIn('replication_status', actual_model_update)
        self.driver._check_volume_type.assert_called_once_with(
            volume, self.fake_nfs_export_1, test_file, {})

    def test_manage_existing_move_fails(self):
        self.mock_object(utils, 'get_file_size', return_value=1074253824)
        self.driver._mounted_shares = [self.fake_nfs_export_1]
        test_file = 'test_file_name'
        volume = fake.FAKE_MANAGE_VOLUME
        vol_path = "%s/%s" % (self.fake_nfs_export_1, test_file)
        vol_ref = {'source-name': vol_path}
        self.driver._check_volume_type = mock.Mock()
        self.driver._ensure_shares_mounted = mock.Mock()
        self.driver._get_mount_point_for_share = mock.Mock(
            return_value=self.fake_mount_point)
        self.driver._get_share_mount_and_vol_from_vol_ref = mock.Mock(
            return_value=(self.fake_nfs_export_1, self.fake_mount_point,
                          test_file))
        self.driver._execute = mock.Mock(
            side_effect=processutils.ProcessExecutionError)
        mock_get_specs = self.mock_object(na_utils, 'get_volume_extra_specs')
        mock_get_specs.return_value = {}
        self.mock_object(self.driver, '_do_qos_for_volume')

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.manage_existing, volume, vol_ref)

    def test_unmanage(self):
        mock_log = self.mock_object(nfs_base, 'LOG')
        volume = {'id': '123', 'provider_location': '/share'}

        retval = self.driver.unmanage(volume)

        self.assertIsNone(retval)
        self.assertEqual(1, mock_log.info.call_count)

    def test_manage_existing_get_size(self):
        test_file = 'test_file_name'
        self.driver._get_share_mount_and_vol_from_vol_ref = mock.Mock(
            return_value=(self.fake_nfs_export_1, self.fake_mount_point,
                          test_file))
        self.mock_object(utils, 'get_file_size', return_value=1073741824)
        self.driver._mounted_shares = [self.fake_nfs_export_1]
        volume = fake.FAKE_MANAGE_VOLUME
        vol_path = "%s/%s" % (self.fake_nfs_export_1, test_file)
        vol_ref = {'source-name': vol_path}

        self.driver._ensure_shares_mounted = mock.Mock()
        self.driver._get_mount_point_for_share = mock.Mock(
            return_value=self.fake_mount_point)

        vol_size = self.driver.manage_existing_get_size(volume, vol_ref)

        self.assertEqual(1, vol_size)

    def test_manage_existing_get_size_round_up(self):
        test_file = 'test_file_name'
        self.driver._get_share_mount_and_vol_from_vol_ref = mock.Mock(
            return_value=(self.fake_nfs_export_1, self.fake_mount_point,
                          test_file))
        self.mock_object(utils, 'get_file_size', return_value=1073760270)
        self.driver._mounted_shares = [self.fake_nfs_export_1]
        volume = fake.FAKE_MANAGE_VOLUME
        vol_path = "%s/%s" % (self.fake_nfs_export_1, test_file)
        vol_ref = {'source-name': vol_path}

        self.driver._ensure_shares_mounted = mock.Mock()
        self.driver._get_mount_point_for_share = mock.Mock(
            return_value=self.fake_mount_point)

        vol_size = self.driver.manage_existing_get_size(volume, vol_ref)

        self.assertEqual(2, vol_size)

    def test_manage_existing_get_size_error(self):
        test_file = 'test_file_name'
        self.driver._get_share_mount_and_vol_from_vol_ref = mock.Mock(
            return_value=(self.fake_nfs_export_1, self.fake_mount_point,
                          test_file))
        self.driver._mounted_shares = [self.fake_nfs_export_1]
        volume = fake.FAKE_MANAGE_VOLUME
        vol_path = "%s/%s" % (self.fake_nfs_export_1, test_file)
        vol_ref = {'source-name': vol_path}

        self.driver._ensure_shares_mounted = mock.Mock()
        self.driver._get_mount_point_for_share = mock.Mock(
            return_value=self.fake_mount_point)

        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.manage_existing_get_size,
                          volume,
                          vol_ref)

    @ddt.data(True, False)
    def test_delete_file(self, volume_not_present):
        mock_get_provider_location = self.mock_object(
            self.driver, '_get_provider_location')
        mock_get_provider_location.return_value = fake.NFS_SHARE
        mock_volume_not_present = self.mock_object(
            self.driver, '_volume_not_present')
        mock_volume_not_present.return_value = volume_not_present
        mock_get_volume_path = self.mock_object(
            self.driver, '_get_volume_path')
        mock_get_volume_path.return_value = fake.PATH
        mock_delete = self.mock_object(self.driver, '_delete')

        self.driver._delete_file(fake.CG_VOLUME_ID, fake.CG_VOLUME_NAME)

        mock_get_provider_location.assert_called_once_with(fake.CG_VOLUME_ID)
        mock_volume_not_present.assert_called_once_with(
            fake.NFS_SHARE, fake.CG_VOLUME_NAME)
        if not volume_not_present:
            mock_get_volume_path.assert_called_once_with(
                fake.NFS_SHARE, fake.CG_VOLUME_NAME)
            mock_delete.assert_called_once_with(fake.PATH)

    def test_delete_file_volume_not_present(self):
        mock_get_provider_location = self.mock_object(
            self.driver, '_get_provider_location')
        mock_get_provider_location.return_value = fake.NFS_SHARE
        mock_volume_not_present = self.mock_object(
            self.driver, '_volume_not_present')
        mock_volume_not_present.return_value = True
        mock_get_volume_path = self.mock_object(
            self.driver, '_get_volume_path')
        mock_delete = self.mock_object(self.driver, '_delete')

        self.driver._delete_file(fake.CG_VOLUME_ID, fake.CG_VOLUME_NAME)

        mock_get_provider_location.assert_called_once_with(fake.CG_VOLUME_ID)
        mock_volume_not_present.assert_called_once_with(
            fake.NFS_SHARE, fake.CG_VOLUME_NAME)
        mock_get_volume_path.assert_not_called()
        mock_delete.assert_not_called()

    def test_check_for_setup_error(self):
        super_check_for_setup_error = self.mock_object(
            nfs.NfsDriver, 'check_for_setup_error')
        mock_start_tasks = self.mock_object(
            self.driver.loopingcalls, 'start_tasks')

        self.driver.check_for_setup_error()

        super_check_for_setup_error.assert_called_once_with()
        mock_start_tasks.assert_called_once_with()

    def test_add_looping_tasks(self):
        mock_add_task = self.mock_object(self.driver.loopingcalls, 'add_task')
        mock_call_snap_cleanup = self.mock_object(
            self.driver, '_delete_snapshots_marked_for_deletion')
        mock_call_ems_logging = self.mock_object(
            self.driver, '_handle_ems_logging')

        self.driver._add_looping_tasks()

        mock_add_task.assert_has_calls([
            mock.call(mock_call_snap_cleanup, loopingcalls.ONE_MINUTE,
                      loopingcalls.ONE_MINUTE),
            mock.call(mock_call_ems_logging, loopingcalls.ONE_HOUR)])

    def test__clone_from_cache(self):
        image_id = 'fake_image_id'
        cache_result = [
            ('fakepool_bad1', '/fakepath/img-cache-1'),
            ('fakepool', '/fakepath/img-cache-2'),
            ('fakepool_bad2', '/fakepath/img-cache-3'),
        ]
        mock_call__is_share_clone_compatible = self.mock_object(
            self.driver, '_is_share_clone_compatible')
        mock_call__is_share_clone_compatible.return_value = True
        mock_call__do_clone_rel_img_cache = self.mock_object(
            self.driver, '_do_clone_rel_img_cache')
        cloned = self.driver._clone_from_cache(fake.test_volume, image_id,
                                               cache_result)
        self.assertTrue(cloned)
        mock_call__is_share_clone_compatible.assert_called_once_with(
            fake.test_volume, 'fakepool')
        mock_call__do_clone_rel_img_cache.assert_called_once_with(
            '/fakepath/img-cache-2', 'fakename', 'fakepool',
            '/fakepath/img-cache-2'
        )

    def test__clone_from_cache_not_found(self):
        image_id = 'fake_image_id'
        cache_result = [
            ('fakepool_bad1', '/fakepath/img-cache-1'),
            ('fakepool_bad2', '/fakepath/img-cache-2'),
            ('fakepool_bad3', '/fakepath/img-cache-3'),
        ]
        mock_call__is_share_clone_compatible = self.mock_object(
            self.driver, '_is_share_clone_compatible')
        mock_call__do_clone_rel_img_cache = self.mock_object(
            self.driver, '_do_clone_rel_img_cache')
        cloned = self.driver._clone_from_cache(fake.test_volume, image_id,
                                               cache_result)
        self.assertFalse(cloned)
        mock_call__is_share_clone_compatible.assert_not_called()
        mock_call__do_clone_rel_img_cache.assert_not_called()

    def test__find_share(self):
        mock_extract = self.mock_object(volume_utils, 'extract_host',
                                        return_value=fake.POOL_NAME)

        pool_name = self.driver._find_share(fake.VOLUME)

        self.assertEqual(pool_name, fake.POOL_NAME)
        mock_extract.assert_called_once_with(fake.VOLUME['host'],
                                             level='pool')

    def test__find_share_error(self):
        mock_extract = self.mock_object(volume_utils, 'extract_host',
                                        return_value=None)

        self.assertRaises(exception.InvalidHost,
                          self.driver._find_share,
                          fake.VOLUME)
        mock_extract.assert_called_once_with(fake.VOLUME['host'],
                                             level='pool')

    def test__ensure_flexgroup_not_in_cg_raises(self):
        self.mock_object(self.driver, '_is_flexgroup',
                         return_value=True)
        self.mock_object(volume_utils, 'is_group_a_cg_snapshot_type',
                         return_value=True)
        fake_v1 = {
            'group': 'fake_group',
            'host': 'fake_host',
            'id': 'fake_id'
        }
        self.assertRaises(na_utils.NetAppDriverException,
                          self.driver._ensure_flexgroup_not_in_cg,
                          fake_v1)

    def test__is_flexgroup_clone_file_supported(self):
        self.assertRaises(NotImplementedError,
                          self.driver._is_flexgroup_clone_file_supported)

    def test_update_migrated_volume(self):
        self.assertRaises(NotImplementedError,
                          self.driver.update_migrated_volume, self.ctxt,
                          fake.test_volume, mock.sentinel.new_volume,
                          mock.sentinel.original_status)

    @ddt.data({'is_flexgroup': False, 'is_flexgroup_supported': False},
              {'is_flexgroup': False, 'is_flexgroup_supported': True},
              {'is_flexgroup': True, 'is_flexgroup_supported': False},
              {'is_flexgroup': True, 'is_flexgroup_supported': True})
    @ddt.unpack
    def test_revert_to_snapshot(self, is_flexgroup, is_flexgroup_supported):
        context = {}

        self.mock_object(self.driver,
                         '_is_flexgroup',
                         return_value=is_flexgroup)
        self.mock_object(self.driver,
                         '_is_flexgroup_clone_file_supported',
                         return_value=is_flexgroup_supported)
        mock_revert_to_snapshot = self.mock_object(
            remotefs.RemoteFSSnapDriverDistributed, 'revert_to_snapshot')
        mock__revert_to_snapshot = self.mock_object(self.driver,
                                                    '_revert_to_snapshot')

        self.driver.revert_to_snapshot(context, fake.SNAPSHOT_VOLUME,
                                       fake.SNAPSHOT)

        if is_flexgroup and not is_flexgroup_supported:
            mock_revert_to_snapshot.assert_called_once_with(
                context, fake.SNAPSHOT_VOLUME, fake.SNAPSHOT)
            mock__revert_to_snapshot.assert_not_called()
        else:
            mock__revert_to_snapshot.assert_called_once_with(
                fake.SNAPSHOT_VOLUME, fake.SNAPSHOT)
            mock_revert_to_snapshot.assert_not_called()

    def test__revert_to_snapshot(self):
        self.assertRaises(NotImplementedError, self.driver._revert_to_snapshot,
                          fake.SNAPSHOT_VOLUME, fake.SNAPSHOT)



# Copyright (c) 2022 DataCore Software Corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Base Driver for DataCore SANsymphony storage array."""

import math
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import loopingcall
from oslo_utils import excutils
from oslo_utils import units

from cinder import context as cinder_context
from cinder import exception as cinder_exception
from cinder.i18n import _
from cinder import interface
from cinder import utils as cinder_utils
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume.drivers.datacore import api
from cinder.volume.drivers.datacore import exception as datacore_exception
from cinder.volume.drivers.datacore import utils as datacore_utils
from cinder.volume.drivers.san import san
from cinder.volume import volume_types

LOG = logging.getLogger(__name__)

datacore_opts = [
    cfg.StrOpt('datacore_disk_type',
               default='single',
               choices=['single', 'mirrored'],
               help='DataCore virtual disk type (single/mirrored). '
                    'Mirrored virtual disks require two storage servers in '
                    'the server group.'),
    cfg.StrOpt('datacore_storage_profile',
               default=None,
               help='DataCore virtual disk storage profile.'),
    cfg.ListOpt('datacore_disk_pools',
                default=[],
                help='List of DataCore disk pools that can be used '
                     'by volume driver.'),
    cfg.IntOpt('datacore_api_timeout',
               default=300,
               min=1,
               help='Seconds to wait for a response from a '
                    'DataCore API call.'),
    cfg.IntOpt('datacore_disk_failed_delay',
               default=300,
               min=0,
               help='Seconds to wait for DataCore virtual '
                    'disk to come out of the "Failed" state.'),
]

CONF = cfg.CONF
CONF.register_opts(datacore_opts, group=configuration.SHARED_CONF_GROUP)


@interface.volumedriver
class DataCoreVolumeDriver(driver.VolumeDriver):
    """DataCore SANsymphony base volume driver."""

    STORAGE_PROTOCOL = 'N/A'

    AWAIT_DISK_ONLINE_INTERVAL = 10
    AWAIT_SNAPSHOT_ONLINE_INTERVAL = 10
    AWAIT_SNAPSHOT_ONLINE_INITIAL_DELAY = 5

    DATACORE_SINGLE_DISK = 'single'
    DATACORE_MIRRORED_DISK = 'mirrored'
    VOLUME_TYPE_STRIPED = 'Striped'
    VOLUME_TYPE_SPANNED = 'Spanned'

    DATACORE_DISK_TYPE_KEY = 'datacore:disk_type'
    DATACORE_STORAGE_PROFILE_KEY = 'datacore:storage_profile'
    DATACORE_DISK_POOLS_KEY = 'datacore:disk_pools'

    VALID_VOLUME_TYPE_KEYS = (DATACORE_DISK_TYPE_KEY,
                              DATACORE_STORAGE_PROFILE_KEY,
                              DATACORE_DISK_POOLS_KEY,)

    def __init__(self, *args, **kwargs):
        super(DataCoreVolumeDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(san.san_opts)
        self.configuration.append_config_values(datacore_opts)
        self._api = None
        self._default_volume_options = None

    def do_setup(self, context):
        """Perform validations and establish connection to server.

        :param context: Context information
        """

        required_params = [
            'san_ip',
            'san_login',
            'san_password',
        ]
        for param in required_params:
            if not getattr(self.configuration, param, None):
                raise cinder_exception.InvalidInput(_("%s not set.") % param)

        self._api = api.DataCoreClient(
            self.configuration.san_ip,
            self.configuration.san_login,
            self.configuration.san_password,
            self.configuration.datacore_api_timeout)

        disk_type = self.configuration.datacore_disk_type
        if disk_type:
            disk_type = disk_type.lower()
        storage_profile = self.configuration.datacore_storage_profile
        if storage_profile:
            storage_profile = storage_profile.lower()
        disk_pools = self.configuration.datacore_disk_pools
        if disk_pools:
            disk_pools = [pool.lower() for pool in disk_pools]

        self._default_volume_options = {
            self.DATACORE_DISK_TYPE_KEY: disk_type,
            self.DATACORE_STORAGE_PROFILE_KEY: storage_profile,
            self.DATACORE_DISK_POOLS_KEY: disk_pools,
        }

    def check_for_setup_error(self):
        pass

    def get_volume_backend_name(self):
        """Get volume backend name of the volume service.

        :return: Volume backend name
        """

        backend_name = self.configuration.safe_get('volume_backend_name')
        return (backend_name or 'DataCore' + self.__class__.__name__)

    def create_volume(self, volume):
        """Creates a volume.

        :param volume: Volume object
        :return: Dictionary of changes to the volume object to be persisted
        """

        volume_options = self._get_volume_options(volume)

        disk_type = volume_options[self.DATACORE_DISK_TYPE_KEY]
        if disk_type == self.DATACORE_MIRRORED_DISK:
            logical_disk_count = 2
            virtual_disk_type = 'MultiPathMirrored'
        elif disk_type == self.DATACORE_SINGLE_DISK:
            logical_disk_count = 1
            virtual_disk_type = 'NonMirrored'
        else:
            msg = _("Virtual disk type '%s' is not valid.") % disk_type
            LOG.error(msg)
            raise cinder_exception.VolumeDriverException(message=msg)

        profile_id = self._get_storage_profile_id(
            volume_options[self.DATACORE_STORAGE_PROFILE_KEY])

        pools = datacore_utils.get_distinct_by(
            lambda pool: pool.ServerId,
            self._get_available_disk_pools(
                volume_options[self.DATACORE_DISK_POOLS_KEY]))

        if len(pools) < logical_disk_count:
            msg = _("Suitable disk pools were not found for "
                    "creating virtual disk.")
            LOG.error(msg)
            raise cinder_exception.VolumeDriverException(message=msg)

        disk_size = self._get_size_in_bytes(volume.size)

        logical_disks = []
        virtual_disk = None
        try:
            for logical_disk_pool in pools[:logical_disk_count]:
                logical_disks.append(
                    self._api.create_pool_logical_disk(
                        logical_disk_pool.Id,
                        self.VOLUME_TYPE_STRIPED, disk_size))

            virtual_disk_data = self._api.build_virtual_disk_data(
                volume.id,
                virtual_disk_type,
                disk_size,
                volume.display_name,
                profile_id)

            virtual_disk = self._api.create_virtual_disk_ex2(
                virtual_disk_data,
                logical_disks[0].Id,
                logical_disks[1].Id if logical_disk_count == 2 else None,
                True)

            virtual_disk = self._await_virtual_disk_online(virtual_disk.Id)

        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Creation of volume %(volume)s failed.",
                              {'volume': volume.id})
                try:
                    if virtual_disk:
                        self._api.delete_virtual_disk(virtual_disk.Id, True)
                    else:
                        for logical_disk in logical_disks:
                            self._api.delete_logical_disk(logical_disk.Id)
                except datacore_exception.DataCoreException as e:
                    LOG.warning("An error occurred on a cleanup after failed "
                                "creation of volume %(volume)s: %(error)s.",
                                {'volume': volume.id, 'error': e})

        return {'provider_location': virtual_disk.Id}

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot.

        :param volume: Volume object
        :param snapshot: Snapshot object
        :return: Dictionary of changes to the volume object to be persisted
        """
        return self._create_volume_from(volume, snapshot)

    def create_cloned_volume(self, volume, src_vref):
        """Creates volume clone.

        :param volume: New Volume object
        :param src_vref: Volume object that must be cloned
        :return: Dictionary of changes to the volume object to be persisted
        """

        return self._create_volume_from(volume, src_vref)

    def extend_volume(self, volume, new_size):
        """Extend an existing volume's size.

        :param volume: Volume object
        :param new_size: new size in GB to extend this volume to
        """

        virtual_disk = self._get_virtual_disk_for(volume, raise_not_found=True)
        self._set_virtual_disk_size(virtual_disk,
                                    self._get_size_in_bytes(new_size))
        virtual_disk = self._await_virtual_disk_online(virtual_disk.Id)

    def delete_volume(self, volume):
        """Deletes a volume.

        :param volume: Volume object
        """

        virtual_disk = self._get_virtual_disk_for(volume)
        if virtual_disk:
            if virtual_disk.IsServed:
                logical_disks = self._api.get_logical_disks()
                logical_units = self._api.get_logical_units()
                target_devices = self._api.get_target_devices()
                logical_disks = [disk.Id for disk in logical_disks
                                 if disk.VirtualDiskId == virtual_disk.Id]
                logical_unit_devices = [unit.VirtualTargetDeviceId
                                        for unit in logical_units
                                        if unit.LogicalDiskId in logical_disks]
                initiator_ports = set(device.InitiatorPortId
                                      for device in target_devices
                                      if device.Id in logical_unit_devices)
                for port in initiator_ports:
                    self._api.unserve_virtual_disks_from_port(
                        port, [virtual_disk.Id])
            self._api.delete_virtual_disk(virtual_disk.Id, True)

    def create_snapshot(self, snapshot):
        """Creates a snapshot.

        :param snapshot: Snapshot object
        :return: Dictionary of changes to the snapshot object to be persisted
        """

        src_virtual_disk = self._get_virtual_disk_for(snapshot.volume,
                                                      raise_not_found=True)

        volume_options = self._get_volume_options(snapshot.volume)
        profile_name = volume_options[self.DATACORE_STORAGE_PROFILE_KEY]
        profile_id = self._get_storage_profile_id(profile_name)
        pool_names = volume_options[self.DATACORE_DISK_POOLS_KEY]

        if src_virtual_disk.DiskStatus != 'Online':
            LOG.warning("Attempting to make a snapshot from virtual disk "
                        "%(disk)s that is in %(state)s state.",
                        {'disk': src_virtual_disk.Id,
                         'state': src_virtual_disk.DiskStatus})

        snapshot_virtual_disk = self._create_virtual_disk_copy(
            src_virtual_disk,
            snapshot.id,
            snapshot.display_name,
            profile_id=profile_id,
            pool_names=pool_names)
        snapshot_virtual_disk = self._await_virtual_disk_online(
            snapshot_virtual_disk.Id)

        return {'provider_location': snapshot_virtual_disk.Id}

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot.

        :param snapshot: Snapshot object
        """

        snapshot_virtual_disk = self._get_virtual_disk_for(snapshot)
        if snapshot_virtual_disk:
            self._api.delete_virtual_disk(snapshot_virtual_disk.Id, True)

    def ensure_export(self, context, volume):
        pass

    def create_export(self, context, volume, connector):
        pass

    def remove_export(self, context, volume):
        pass

    def unserve_virtual_disks_from_host(self, volume, connector):
        virtual_disk = self._get_virtual_disk_for(volume)
        if virtual_disk:
            if connector is None:
                clients = self._api.get_clients()
            else:
                clients = [self._get_client(connector['host'],
                                            create_new=False)]

            server_group = self._get_our_server_group()

            @cinder_utils.synchronized(
                'datacore-backend-%s' % server_group.Id, external=True)
            def unserve_virtual_disk(client_id):
                self._api.unserve_virtual_disks_from_host(
                    client_id, [virtual_disk.Id])

            for client in clients:
                unserve_virtual_disk(client.Id)

    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector.

        :param volume: Volume object
        :param connector: Connector information
        """
        self.unserve_virtual_disks_from_host(volume, connector)

    def manage_existing(self, volume, existing_ref):
        return self.manage_existing_object(volume, existing_ref, "volume")

    def manage_existing_get_size(self, volume, existing_ref):
        return self.manage_existing_object_get_size(volume, existing_ref,
                                                    "volume")

    def manage_existing_snapshot(self, snapshot, existing_ref):
        return self.manage_existing_object(snapshot, existing_ref, "snapshot")

    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        return self.manage_existing_object_get_size(snapshot, existing_ref,
                                                    "snapshot")

    def manage_existing_object(self, existing_object, existing_ref,
                               object_type):
        if 'source-name' not in existing_ref:
            reason = _('Reference must contain source-name element.')
            raise cinder_exception.ManageExistingInvalidReference(
                existing_ref=existing_ref, reason=reason)

        vd_alias = existing_ref['source-name']
        virtual_disk = datacore_utils.get_first_or_default(
            lambda disk: disk.Alias == vd_alias,
            self._api.get_virtual_disks(),
            None)

        if not virtual_disk:
            kwargs = {'existing_ref': vd_alias,
                      'reason': 'Specified Virtual disk does not exist.'}
            raise cinder_exception.ManageExistingInvalidReference(**kwargs)

        return {'provider_location': virtual_disk.Id}

    def manage_existing_object_get_size(self, existing_object, existing_ref,
                                        object_type):
        if 'source-name' not in existing_ref:
            reason = _('Reference must contain source-name element.')
            raise cinder_exception.ManageExistingInvalidReference(
                existing_ref=existing_ref, reason=reason)

        vd_alias = existing_ref['source-name']
        virtual_disk = datacore_utils.get_first_or_default(
            lambda disk: disk.Alias == vd_alias,
            self._api.get_virtual_disks(),
            None)

        if not virtual_disk:
            kwargs = {'existing_ref': vd_alias,
                      'reason': 'Specified Virtual disk does not exist.'}
            raise cinder_exception.ManageExistingInvalidReference(**kwargs)
        return(self._get_size_in_gigabytes(virtual_disk.Size.Value))

    def _update_volume_stats(self):
        performance_data = self._api.get_performance_by_type(
            ['DiskPoolPerformance'])
        total = 0
        available = 0
        reserved = 0
        for performance in performance_data:
            missing_perf_data = []

            if hasattr(performance.PerformanceData, 'BytesTotal'):
                total += performance.PerformanceData.BytesTotal
            else:
                missing_perf_data.append('BytesTotal')

            if hasattr(performance.PerformanceData, 'BytesAvailable'):
                available += performance.PerformanceData.BytesAvailable
            else:
                missing_perf_data.append('BytesAvailable')

            if hasattr(performance.PerformanceData, 'BytesReserved'):
                reserved += performance.PerformanceData.BytesReserved
            else:
                missing_perf_data.append('BytesReserved')

            if missing_perf_data:
                LOG.warning("Performance data %(data)s is missing for "
                            "disk pool %(pool)s",
                            {'data': missing_perf_data,
                             'pool': performance.ObjectId})
        provisioned = 0
        logical_disks = self._api.get_logical_disks()
        for disk in logical_disks:
            if getattr(disk, 'PoolId', None):
                provisioned += disk.Size.Value
        total_capacity_gb = self._get_size_in_gigabytes(total)
        free = available + reserved
        free_capacity_gb = self._get_size_in_gigabytes(free)
        provisioned_capacity_gb = self._get_size_in_gigabytes(provisioned)
        reserved_percentage = 100.0 * reserved / total if total else 0.0
        reserved_percentage = math.ceil(reserved_percentage)
        ratio = self.configuration.max_over_subscription_ratio
        stats_data = {
            'vendor_name': 'DataCore',
            'QoS_support': False,
            'volume_backend_name': self.get_volume_backend_name(),
            'driver_version': self.get_version(),
            'storage_protocol': self.STORAGE_PROTOCOL,
            'total_capacity_gb': total_capacity_gb,
            'free_capacity_gb': free_capacity_gb,
            'provisioned_capacity_gb': provisioned_capacity_gb,
            'reserved_percentage': reserved_percentage,
            'max_over_subscription_ratio': ratio,
            'thin_provisioning_support': True,
            'thick_provisioning_support': False,
            'online_extend_support': False,
        }
        self._stats = stats_data

    def _get_our_server_group(self):
        server_group = datacore_utils.get_first(lambda group: group.OurGroup,
                                                self._api.get_server_groups())

        return server_group

    def _get_volume_options_from_type(self, type_id, default_options):
        options = dict(default_options.items())
        if type_id:
            admin_context = cinder_context.get_admin_context()
            volume_type = volume_types.get_volume_type(admin_context, type_id)
            specs = dict(volume_type).get('extra_specs')

            for key, value in specs.items():
                if key in self.VALID_VOLUME_TYPE_KEYS:
                    if key == self.DATACORE_DISK_POOLS_KEY:
                        options[key] = [v.strip().lower()
                                        for v in value.split(',')]
                    else:
                        options[key] = value.lower()

        return options

    def _get_volume_options(self, volume):
        type_id = volume.volume_type_id

        volume_options = self._get_volume_options_from_type(
            type_id, self._default_volume_options)

        return volume_options

    def _get_online_servers(self):
        servers = self._api.get_servers()
        online_servers = [server for server in servers
                          if server.State == 'Online']
        return online_servers

    def _get_available_disk_pools(self, disk_pool_names=None):
        online_servers = [server.Id for server in self._get_online_servers()]

        pool_performance = {
            performance.ObjectId: performance.PerformanceData for performance
            in self._api.get_performance_by_type(['DiskPoolPerformance'])}

        disk_pools = self._api.get_disk_pools()

        lower_disk_pool_names = ([name.lower() for name in disk_pool_names]
                                 if disk_pool_names else [])

        available_disk_pools = [
            pool for pool in disk_pools
            if (self._is_pool_healthy(pool, pool_performance,
                                      online_servers) and
                (not lower_disk_pool_names or
                 pool.Caption.lower() in lower_disk_pool_names))]

        available_disk_pools.sort(
            key=lambda p: pool_performance[p.Id].BytesAvailable, reverse=True)

        return available_disk_pools

    def _get_virtual_disk_for(self, obj, raise_not_found=False):
        disk_id = obj.get('provider_location')

        virtual_disk = datacore_utils.get_first_or_default(
            lambda disk: disk.Id == disk_id,
            self._api.get_virtual_disks(),
            None)
        if not virtual_disk:
            msg = (_("Virtual disk not found for %(object)s %(object_id)s.")
                   % {'object': obj.__class__.__name__.lower(),
                      'object_id': obj['id']})
            if raise_not_found:
                LOG.error(msg)
                raise cinder_exception.VolumeDriverException(message=msg)
            else:
                LOG.warning(msg)

        return virtual_disk

    def _set_virtual_disk_size(self, virtual_disk, new_size):
        return self._api.set_virtual_disk_size(virtual_disk.Id, new_size)

    def _get_storage_profile(self, profile_name, raise_not_found=False):
        profiles = self._api.get_storage_profiles()
        profile = datacore_utils.get_first_or_default(
            lambda p: p.Caption.lower() == profile_name.lower(),
            profiles,
            None)
        if not profile and raise_not_found:
            msg = (_("Specified storage profile %s not found.")
                   % profile_name)
            LOG.error(msg)
            raise cinder_exception.VolumeDriverException(message=msg)

        return profile

    def _get_storage_profile_id(self, profile_name):
        profile_id = None
        if profile_name:
            profile = self._get_storage_profile(profile_name,
                                                raise_not_found=True)
            profile_id = profile.Id
        return profile_id

    def _await_virtual_disk_online(self, virtual_disk_id):
        def inner(start_time):
            disk_failed_delay = self.configuration.datacore_disk_failed_delay
            virtual_disk = datacore_utils.get_first(
                lambda disk: disk.Id == virtual_disk_id,
                self._api.get_virtual_disks())
            if virtual_disk.DiskStatus == 'Online':
                raise loopingcall.LoopingCallDone(virtual_disk)
            elif (
                    virtual_disk.DiskStatus != 'FailedRedundancy' and
                    time.time() - start_time >= disk_failed_delay):
                msg = (_("Virtual disk %(disk)s did not come out of the "
                         "%(state)s state after %(timeout)s seconds.")
                       % {'disk': virtual_disk.Id,
                          'state': virtual_disk.DiskStatus,
                          'timeout': disk_failed_delay})
                LOG.error(msg)
                raise cinder_exception.VolumeDriverException(message=msg)

        inner_loop = loopingcall.FixedIntervalLoopingCall(inner, time.time())
        return inner_loop.start(self.AWAIT_DISK_ONLINE_INTERVAL).wait()

    def _create_volume_from(self, volume, src_obj):
        src_virtual_disk = self._get_virtual_disk_for(src_obj,
                                                      raise_not_found=True)
        if src_virtual_disk.DiskStatus != 'Online':
            LOG.warning("Attempting to create a volume from virtual disk "
                        "%(disk)s that is in %(state)s state.",
                        {'disk': src_virtual_disk.Id,
                         'state': src_virtual_disk.DiskStatus})

        volume_options = self._get_volume_options(volume)
        profile_id = self._get_storage_profile_id(
            volume_options[self.DATACORE_STORAGE_PROFILE_KEY])
        pool_names = volume_options[self.DATACORE_DISK_POOLS_KEY]

        volume_virtual_disk = self._create_virtual_disk_copy(
            src_virtual_disk,
            volume.id,
            volume.display_name,
            profile_id=profile_id,
            pool_names=pool_names)

        volume_logical_disk = datacore_utils.get_first(
            lambda disk: disk.VirtualDiskId == volume_virtual_disk.Id,
            self._api.get_logical_disks())

        try:
            disk_type = volume_options[self.DATACORE_DISK_TYPE_KEY]
            if disk_type == self.DATACORE_MIRRORED_DISK:
                pools = self._get_available_disk_pools(pool_names)
                selected_pool = datacore_utils.get_first_or_default(
                    lambda pool: (pool.ServerId !=
                                  volume_logical_disk.ServerHostId and
                                  pool.Id != volume_logical_disk.PoolId),
                    pools, None)
                if selected_pool:
                    logical_disk = self._api.create_pool_logical_disk(
                        selected_pool.Id,
                        self.VOLUME_TYPE_STRIPED,
                        volume_virtual_disk.Size.Value)
                    self._api.bind_logical_disk(volume_virtual_disk.Id,
                                                logical_disk.Id,
                                                'Second',
                                                True,
                                                False,
                                                True)
                else:
                    msg = _("Can not create mirrored virtual disk. "
                            "Suitable disk pools not found.")
                    LOG.error(msg)
                    raise cinder_exception.VolumeDriverException(message=msg)
            volume_virtual_disk = self._await_virtual_disk_online(
                volume_virtual_disk.Id)
            try:
                source_size = src_obj.size
            except AttributeError:
                source_size = src_obj.volume_size
            if volume.size > source_size:
                self._set_virtual_disk_size(volume_virtual_disk,
                                            self._get_size_in_bytes(
                                                volume.size))
                volume_virtual_disk = datacore_utils.get_first(
                    lambda disk: disk.Id == volume_virtual_disk.Id,
                    self._api.get_virtual_disks())
                volume_virtual_disk = self._await_virtual_disk_size_change(
                    volume_virtual_disk.Id,
                    self._get_size_in_bytes(source_size))

        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Creation of volume %(volume)s failed.",
                              {'volume': volume.id})
                try:
                    self._api.delete_virtual_disk(volume_virtual_disk.Id, True)
                except datacore_exception.DataCoreException as e:
                    LOG.warning("An error occurred on a cleanup after failed "
                                "creation of volume %(volume)s: %(error)s.",
                                {'volume': volume.id, 'error': e})

        return {'provider_location': volume_virtual_disk.Id}

    def _create_full_snapshot(self, description, name, pool_names, profile_id,
                              src_virtual_disk):
        pools = self._get_available_disk_pools(pool_names)
        destination_pool = datacore_utils.get_first_or_default(
            lambda pool: (pool.ServerId == src_virtual_disk.FirstHostId or
                          pool.ServerId == src_virtual_disk.SecondHostId),
            pools, None)

        if not destination_pool:
            msg = _("Suitable snapshot destination disk pool not found for "
                    "virtual disk %s.") % src_virtual_disk.Id
            LOG.error(msg)
            raise cinder_exception.VolumeDriverException(message=msg)
        server = datacore_utils.get_first(
            lambda srv: srv.Id == destination_pool.ServerId,
            self._api.get_servers())
        if not server.SnapshotMapStorePoolId:
            self._api.designate_map_store(destination_pool.Id)
        snapshot = self._api.create_snapshot(src_virtual_disk.Id,
                                             name,
                                             description,
                                             destination_pool.Id,
                                             'Full',
                                             False,
                                             profile_id)
        return snapshot

    def _await_snapshot_migrated(self, snapshot_id):
        def inner():
            snapshot_data = datacore_utils.get_first(
                lambda snapshot: snapshot.Id == snapshot_id,
                self._api.get_snapshots())
            if snapshot_data.State == 'Migrated':
                raise loopingcall.LoopingCallDone(snapshot_data)
            elif (snapshot_data.State != 'Healthy' and
                  snapshot_data.Failure != 'NoFailure'):
                msg = (_("Full migration of snapshot %(snapshot)s failed. "
                         "Snapshot is in %(state)s state.")
                       % {'snapshot': snapshot_data.Id,
                          'state': snapshot_data.State})
                LOG.error(msg)
                raise cinder_exception.VolumeDriverException(message=msg)

        loop = loopingcall.FixedIntervalLoopingCall(inner)
        time.sleep(self.AWAIT_SNAPSHOT_ONLINE_INTERVAL)
        return loop.start(self.AWAIT_SNAPSHOT_ONLINE_INTERVAL,
                          self.AWAIT_SNAPSHOT_ONLINE_INITIAL_DELAY).wait()

    def _create_virtual_disk_copy(self, src_virtual_disk, name, description,
                                  profile_id=None, pool_names=None):
        snapshot = self._create_full_snapshot(
            description, name, pool_names, profile_id, src_virtual_disk)

        try:
            snapshot = self._await_snapshot_migrated(snapshot.Id)
            self._api.delete_snapshot(snapshot.Id)
            self._await_snapshot_split_state_change(snapshot)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Split operation failed for snapshot "
                              "%(snapshot)s.", {'snapshot': snapshot.Id})
                try:
                    logical_disk_copy = datacore_utils.get_first(
                        lambda disk: (disk.Id ==
                                      snapshot.DestinationLogicalDiskId),
                        self._api.get_logical_disks())

                    virtual_disk_copy = datacore_utils.get_first(
                        lambda disk: (disk.Id ==
                                      logical_disk_copy.VirtualDiskId),
                        self._api.get_virtual_disks())

                    self._api.delete_virtual_disk(virtual_disk_copy.Id, True)
                except datacore_exception.DataCoreException as e:
                    LOG.warning("An error occurred on a cleanup after failed "
                                "split of snapshot %(snapshot)s: %(error)s.",
                                {'snapshot': snapshot.Id, 'error': e})

        logical_disk_copy = datacore_utils.get_first(
            lambda disk: disk.Id == snapshot.DestinationLogicalDiskId,
            self._api.get_logical_disks())

        virtual_disk_copy = datacore_utils.get_first(
            lambda disk: disk.Id == logical_disk_copy.VirtualDiskId,
            self._api.get_virtual_disks())

        return virtual_disk_copy

    def _get_client(self, name, create_new=False):
        client_hosts = self._api.get_clients()

        client = datacore_utils.get_first_or_default(
            lambda host: host.HostName.split('.')[0] == name.split('.')[0],
            client_hosts, None)

        if create_new:
            if not client:
                client = self._api.register_client(
                    name, None, 'Other', 'PreferredServer', None)
            self._api.set_client_capabilities(client.Id, True, True)

        return client

    @staticmethod
    def _is_pool_healthy(pool, pool_performance, online_servers):
        if (pool.PoolStatus == 'Running' and
                hasattr(pool_performance[pool.Id], 'BytesAvailable') and
                pool.ServerId in online_servers):
            return True
        return False

    @staticmethod
    def _get_size_in_bytes(size_in_gigabytes):
        return size_in_gigabytes * units.Gi

    @staticmethod
    def _get_size_in_gigabytes(size_in_bytes):
        return size_in_bytes / float(units.Gi)

    def _await_virtual_disk_size_change(self, virtual_disk_id, old_size):
        def inner(start_time):
            disk_failed_delay = self.configuration.datacore_disk_failed_delay
            virtual_disk = datacore_utils.get_first(
                lambda disk: disk.Id == virtual_disk_id,
                self._api.get_virtual_disks())
            if virtual_disk.DiskStatus == 'Online' \
                    and virtual_disk.Size.Value > old_size:
                raise loopingcall.LoopingCallDone(virtual_disk)
            elif (virtual_disk.DiskStatus != 'FailedRedundancy' and
                  time.time() - start_time >= disk_failed_delay):
                msg = (_("Virtual disk %(disk)s did not come out of the "
                         "%(state)s state after %(timeout)s seconds.")
                       % {'disk': virtual_disk.Id,
                          'state': virtual_disk.DiskStatus,
                          'timeout': disk_failed_delay})
                LOG.error(msg)
                raise cinder_exception.VolumeDriverException(message=msg)

        inner_loop = loopingcall.FixedIntervalLoopingCall(inner, time.time())
        time.sleep(self.AWAIT_DISK_ONLINE_INTERVAL)
        return inner_loop.start(self.AWAIT_DISK_ONLINE_INTERVAL).wait()

    def _await_snapshot_split_state_change(self, split_snapshot):
        def inner(start_time):
            disk_failed_delay = self.configuration.datacore_disk_failed_delay
            snapshot_found = False
            snapshot_list = self._api.get_snapshots()
            if not snapshot_list:
                raise loopingcall.LoopingCallDone()
            for entry in snapshot_list:
                if entry.Caption == split_snapshot.Caption:
                    snapshot_found = True
                    break
            if not snapshot_found:
                raise loopingcall.LoopingCallDone()
            if (time.time() - start_time >= disk_failed_delay):
                msg = (_("Split Snapshot disk %(disk)s did not happened "
                         "after %(timeout)s seconds.")
                       % {'disk': split_snapshot.Caption,
                          'timeout': disk_failed_delay})
                LOG.error(msg)
                raise loopingcall.LoopingCallDone()

        inner_loop = loopingcall.FixedIntervalLoopingCall(inner, time.time())
        return inner_loop.start(self.AWAIT_DISK_ONLINE_INTERVAL).wait()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy

import ddt

from cinder.api.contrib import volume_type_access as vta
from cinder.api import microversions as mv
from cinder import objects
from cinder.policies import volume_access as vta_policies
from cinder.tests.unit.api.contrib import test_volume_type_access as vta_test
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit.policies import base

IS_PUBLIC_FIELD = 'os-volume-type-access:is_public'


# the original uses a class var and admin context
class FakeRequest(vta_test.FakeRequest):
    def __init__(self, context):
        self.environ = {"cinder.context": context}


FAKE_RESP_OBJ = {
    'volume_type': {'id': fake.VOLUME_TYPE_ID},
    'volume_types': [
        {'id': fake.VOLUME_TYPE_ID},
        {'id': fake.VOLUME_TYPE3_ID}
    ]}


# need an instance var so this will work with ddt
class FakeResponse(vta_test.FakeResponse):
    def __init__(self):
        self.obj = deepcopy(FAKE_RESP_OBJ)


@ddt.ddt
class VolumeTypeAccessFieldPolicyTest(base.BasePolicyTest):

    # NOTE: We are testing directly against the extension controller.
    # Its call to context.authorize doesn't provide a target, so
    # "is_admin" or "project_id:%(project_id)s" always matches.
    authorized_users = [
        'legacy_admin',
        'project_admin',
        'system_admin',
        'legacy_owner',
        'project_member',
        'project_reader',
        'project_foo',
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # note: authorize is called with fatal=False, so everyone is a winner!
    everyone = authorized_users

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = vta.VolumeTypeActionController()
        self.rule_name = vta_policies.TYPE_ACCESS_POLICY
        self.api_version = mv.BASE_VERSION
        self.api_path = f'/v3/{self.project_id}/types'

    @ddt.data(*base.all_users)
    def test_type_access_policy_types_list(self, user_id):
        unauthorized_exceptions = None
        req = FakeRequest(self.create_context(user_id))
        resp = FakeResponse()

        self.common_policy_check(user_id,
                                 self.everyone,
                                 [],
                                 unauthorized_exceptions,
                                 self.rule_name,
                                 self.controller.index,
                                 req,
                                 resp)

        # this is where the real check happens
        if user_id in self.authorized_users:
            for vol_type in resp.obj['volume_types']:
                self.assertIn(IS_PUBLIC_FIELD, vol_type)
        else:
            for vol_type in resp.obj['volume_types']:
                self.assertNotIn(IS_PUBLIC_FIELD, vol_type)

    @ddt.data(*base.all_users)
    def test_type_access_policy_type_show(self, user_id):
        unauthorized_exceptions = None
        req = FakeRequest(self.create_context(user_id))
        resp = FakeResponse()

        self.common_policy_check(user_id,
                                 self.everyone,
                                 [],
                                 unauthorized_exceptions,
                                 self.rule_name,
                                 self.controller.show,
                                 req,
                                 resp,
                                 fake.VOLUME_TYPE_ID)

        if user_id in self.authorized_users:
            self.assertIn(IS_PUBLIC_FIELD, resp.obj['volume_type'])
        else:
            self.assertNotIn(IS_PUBLIC_FIELD, resp.obj['volume_type'])

    @ddt.data(*base.all_users)
    def test_type_access_policy_type_create(self, user_id):
        unauthorized_exceptions = None
        req = FakeRequest(self.create_context(user_id))
        resp = FakeResponse()
        body = None

        self.common_policy_check(user_id,
                                 self.everyone,
                                 [],
                                 unauthorized_exceptions,
                                 self.rule_name,
                                 self.controller.create,
                                 req,
                                 body,
                                 resp)

        if user_id in self.authorized_users:
            self.assertIn(IS_PUBLIC_FIELD, resp.obj['volume_type'])
        else:
            self.assertNotIn(IS_PUBLIC_FIELD, resp.obj['volume_type'])


class VolumeTypeAccessFieldPolicySecureRbacTest(
        VolumeTypeAccessFieldPolicyTest):

    # Remember that we are testing directly against the extension controller,
    # so while the below may seem over-permissive, in real life there is
    # a more selective check that happens first.
    authorized_users = [
        'legacy_admin',
        'project_admin',
        'system_admin',
        'project_member',
        'system_member',
        'other_project_member',
    ]
    # this will be anyone without the 'admin' or 'member' role
    unauthorized_users = [
        'legacy_owner',
        'project_foo',
        'project_reader',
        'system_reader',
        'system_foo',
        'other_project_reader',
    ]
    everyone = authorized_users + unauthorized_users

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)


@ddt.ddt
class VolumeTypeAccessListProjectsPolicyTest(base.BasePolicyTest):
    authorized_users = [
        'legacy_admin',
        'project_admin',
        'system_admin',
    ]
    unauthorized_users = [
        'legacy_owner',
        'project_member',
        'project_reader',
        'project_foo',
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = vta.VolumeTypeAccessController()
        self.volume_type = objects.VolumeType(
            self.project_admin_context,
            name='private_volume_type',
            is_public=False,
            description='volume type for srbac testing',
            extra_specs=None,
            projects=[self.project_id, self.project_id_other])
        self.volume_type.create()
        self.addCleanup(self.volume_type.destroy)
        self.api_version = mv.BASE_VERSION
        self.api_path = (f'/v3/{self.project_id}/types/'
                         f'{self.volume_type.id}/os-volume-type-access')

    @ddt.data(*base.all_users)
    def test_type_access_who_policy(self, user_id):
        """Test policy for listing projects with access to a volume type."""

        rule_name = vta_policies.TYPE_ACCESS_WHO_POLICY
        unauthorized_exceptions = None
        req = fake_api.HTTPRequest.blank(self.api_path)

        self.common_policy_check(user_id,
                                 self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller.index,
                                 req,
                                 self.volume_type.id)


class VolumeTypeAccessListProjectsPolicySecureRbacTest(
        VolumeTypeAccessListProjectsPolicyTest):

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)



# Copyright 2011 Denali Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from http import HTTPStatus
from unittest import mock
from urllib import parse as urllib

import ddt
from oslo_config import cfg
import pytz
import webob

from cinder.api import common
from cinder.api.v2 import snapshots
from cinder import context
from cinder import db
from cinder import exception
from cinder import objects
from cinder.objects import fields
from cinder.scheduler import rpcapi as scheduler_rpcapi
from cinder.tests.unit.api import fakes
from cinder.tests.unit.api.v2 import fakes as v2_fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.tests.unit import utils
from cinder import volume


CONF = cfg.CONF

UUID = '00000000-0000-0000-0000-000000000001'
INVALID_UUID = '00000000-0000-0000-0000-000000000002'


def _get_default_snapshot_param():
    return {
        'id': UUID,
        'volume_id': fake.VOLUME_ID,
        'status': fields.SnapshotStatus.AVAILABLE,
        'volume_size': 100,
        'created_at': None,
        'updated_at': None,
        'user_id': 'bcb7746c7a41472d88a1ffac89ba6a9b',
        'project_id': '7ffe17a15c724e2aa79fc839540aec15',
        'display_name': 'Default name',
        'display_description': 'Default description',
        'deleted': None,
        'volume': {'availability_zone': 'test_zone'}
    }


def fake_snapshot_delete(self, context, snapshot):
    if snapshot['id'] != UUID:
        raise exception.SnapshotNotFound(snapshot['id'])


def fake_snapshot_get(self, context, snapshot_id):
    if snapshot_id != UUID:
        raise exception.SnapshotNotFound(snapshot_id)

    param = _get_default_snapshot_param()
    return param


def fake_snapshot_get_all(self, context, search_opts=None):
    param = _get_default_snapshot_param()
    return [param]


@ddt.ddt
class SnapshotApiTest(test.TestCase):
    def setUp(self):
        super(SnapshotApiTest, self).setUp()
        self.mock_object(scheduler_rpcapi.SchedulerAPI, 'create_snapshot')
        self.controller = snapshots.SnapshotsController()
        self.ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)

    def test_snapshot_create(self):
        volume = utils.create_volume(self.ctx, volume_type_id=None)
        snapshot_name = 'Snapshot Test Name'
        snapshot_description = 'Snapshot Test Desc'
        snapshot = {
            "volume_id": volume.id,
            "force": False,
            "name": snapshot_name,
            "description": snapshot_description
        }

        body = dict(snapshot=snapshot)
        req = fakes.HTTPRequest.blank('/v3/snapshots')
        resp_dict = self.controller.create(req, body=body)

        self.assertIn('snapshot', resp_dict)
        self.assertEqual(snapshot_name, resp_dict['snapshot']['name'])
        self.assertEqual(snapshot_description,
                         resp_dict['snapshot']['description'])
        self.assertIn('updated_at', resp_dict['snapshot'])
        db.volume_destroy(self.ctx, volume.id)

    def test_snapshot_create_with_null_validate(self):

        volume = utils.create_volume(self.ctx, volume_type_id=None)
        snapshot = {
            "volume_id": volume.id,
            "force": False,
            "name": None,
            "description": None
        }

        body = dict(snapshot=snapshot)
        req = fakes.HTTPRequest.blank('/v3/snapshots')
        resp_dict = self.controller.create(req, body=body)

        self.assertIn('snapshot', resp_dict)
        self.assertIsNone(resp_dict['snapshot']['name'])
        self.assertIsNone(resp_dict['snapshot']['description'])
        db.volume_destroy(self.ctx, volume.id)

    @ddt.data(True, 'y', 'true', 'yes', '1', 'on')
    def test_snapshot_create_force(self, force_param):
        volume = utils.create_volume(self.ctx, status='in-use',
                                     volume_type_id=None)
        snapshot_name = 'Snapshot Test Name'
        snapshot_description = 'Snapshot Test Desc'
        snapshot = {
            "volume_id": volume.id,
            "force": force_param,
            "name": snapshot_name,
            "description": snapshot_description
        }
        body = dict(snapshot=snapshot)
        req = fakes.HTTPRequest.blank('/v3/snapshots')
        resp_dict = self.controller.create(req, body=body)

        self.assertIn('snapshot', resp_dict)
        self.assertEqual(snapshot_name,
                         resp_dict['snapshot']['name'])
        self.assertEqual(snapshot_description,
                         resp_dict['snapshot']['description'])
        self.assertIn('updated_at', resp_dict['snapshot'])

        db.volume_destroy(self.ctx, volume.id)

    @ddt.data(False, 'n', 'false', 'No', '0', 'off')
    def test_snapshot_create_force_failure(self, force_param):
        volume = utils.create_volume(self.ctx, status='in-use',
                                     volume_type_id=None)
        snapshot_name = 'Snapshot Test Name'
        snapshot_description = 'Snapshot Test Desc'
        snapshot = {
            "volume_id": volume.id,
            "force": force_param,
            "name": snapshot_name,
            "description": snapshot_description
        }
        body = dict(snapshot=snapshot)
        req = fakes.HTTPRequest.blank('/v3/snapshots')
        self.assertRaises(exception.InvalidVolume,
                          self.controller.create,
                          req,
                          body=body)

        db.volume_destroy(self.ctx, volume.id)

    @ddt.data("**&&^^%%$$##@@", '-1', 2, '01', 'falSE', 0, 'trUE', 1,
              "1         ")
    def test_snapshot_create_invalid_force_param(self, force_param):
        volume = utils.create_volume(self.ctx, status='available',
                                     volume_type_id=None)
        snapshot_name = 'Snapshot Test Name'
        snapshot_description = 'Snapshot Test Desc'

        snapshot = {
            "volume_id": volume.id,
            "force": force_param,
            "name": snapshot_name,
            "description": snapshot_description
        }
        body = dict(snapshot=snapshot)
        req = fakes.HTTPRequest.blank('/v3/snapshots')
        self.assertRaises(exception.ValidationError,
                          self.controller.create,
                          req,
                          body=body)

        db.volume_destroy(self.ctx, volume.id)

    def test_snapshot_create_without_volume_id(self):
        snapshot_name = 'Snapshot Test Name'
        snapshot_description = 'Snapshot Test Desc'
        body = {
            "snapshot": {
                "force": True,
                "name": snapshot_name,
                "description": snapshot_description
            }
        }
        req = fakes.HTTPRequest.blank('/v3/snapshots')
        self.assertRaises(exception.ValidationError,
                          self.controller.create, req, body=body)

    @ddt.data({"snapshot": {"description": "   sample description",
                            "name": "   test"}},
              {"snapshot": {"description": "sample description   ",
                            "name": "test   "}},
              {"snapshot": {"description": " sample description ",
                            "name": "  test name  "}})
    def test_snapshot_create_with_leading_trailing_spaces(self, body):
        volume = utils.create_volume(self.ctx, volume_type_id=None)
        body['snapshot']['volume_id'] = volume.id
        req = fakes.HTTPRequest.blank('/v3/snapshots')
        resp_dict = self.controller.create(req, body=body)

        self.assertEqual(body['snapshot']['display_name'].strip(),
                         resp_dict['snapshot']['name'])
        self.assertEqual(body['snapshot']['description'].strip(),
                         resp_dict['snapshot']['description'])
        db.volume_destroy(self.ctx, volume.id)

    @mock.patch.object(volume.api.API, "update_snapshot",
                       side_effect=v2_fakes.fake_snapshot_update)
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    @mock.patch('cinder.db.volume_get')
    @mock.patch('cinder.objects.Snapshot.get_by_id')
    def test_snapshot_update(
            self, snapshot_get_by_id, volume_get,
            snapshot_metadata_get, update_snapshot):
        snapshot = {
            'id': UUID,
            'volume_id': fake.VOLUME_ID,
            'status': fields.SnapshotStatus.AVAILABLE,
            'created_at': "2014-01-01 00:00:00",
            'volume_size': 100,
            'display_name': 'Default name',
            'display_description': 'Default description',
            'expected_attrs': ['metadata'],
        }
        snapshot_obj = fake_snapshot.fake_snapshot_obj(self.ctx, **snapshot)
        fake_volume_obj = fake_volume.fake_volume_obj(self.ctx)
        snapshot_get_by_id.return_value = snapshot_obj
        volume_get.return_value = fake_volume_obj

        updates = {
            "name": "Updated Test Name",
        }
        body = {"snapshot": updates}
        req = fakes.HTTPRequest.blank('/v3/snapshots/%s' % UUID)
        req.environ['cinder.context'] = self.ctx
        res_dict = self.controller.update(req, UUID, body=body)
        expected = {
            'snapshot': {
                'id': UUID,
                'volume_id': fake.VOLUME_ID,
                'status': fields.SnapshotStatus.AVAILABLE,
                'size': 100,
                'created_at': datetime.datetime(2014, 1, 1, 0, 0, 0,
                                                tzinfo=pytz.utc),
                'updated_at': None,
                'name': u'Updated Test Name',
                'description': u'Default description',
                'metadata': {},
            }
        }
        self.assertEqual(expected, res_dict)
        self.assertEqual(2, len(self.notifier.notifications))

    @mock.patch.object(volume.api.API, "update_snapshot",
                       side_effect=v2_fakes.fake_snapshot_update)
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    @mock.patch('cinder.db.volume_get')
    @mock.patch('cinder.objects.Snapshot.get_by_id')
    def test_snapshot_update_with_null_validate(
            self, snapshot_get_by_id, volume_get,
            snapshot_metadata_get, update_snapshot):
        snapshot = {
            'id': UUID,
            'volume_id': fake.VOLUME_ID,
            'status': fields.SnapshotStatus.AVAILABLE,
            'created_at': "2014-01-01 00:00:00",
            'volume_size': 100,
            'name': 'Default name',
            'description': 'Default description',
            'expected_attrs': ['metadata'],
        }
        snapshot_obj = fake_snapshot.fake_snapshot_obj(self.ctx, **snapshot)
        fake_volume_obj = fake_volume.fake_volume_obj(self.ctx)
        snapshot_get_by_id.return_value = snapshot_obj
        volume_get.return_value = fake_volume_obj

        updates = {
            "name": None,
            "description": None,
        }
        body = {"snapshot": updates}
        req = fakes.HTTPRequest.blank('/v3/snapshots/%s' % UUID)
        req.environ['cinder.context'] = self.ctx
        res_dict = self.controller.update(req, UUID, body=body)

        self.assertEqual(fields.SnapshotStatus.AVAILABLE,
                         res_dict['snapshot']['status'])
        self.assertIsNone(res_dict['snapshot']['name'])
        self.assertIsNone(res_dict['snapshot']['description'])

    def test_snapshot_update_missing_body(self):
        body = {}
        req = fakes.HTTPRequest.blank('/v3/snapshots/%s' % UUID)
        self.assertRaises(exception.ValidationError,
                          self.controller.update, req, UUID, body=body)

    def test_snapshot_update_invalid_body(self):
        body = {'name': 'missing top level snapshot key'}
        req = fakes.HTTPRequest.blank('/v3/snapshots/%s' % UUID)
        self.assertRaises(exception.ValidationError,
                          self.controller.update, req, UUID, body=body)

    def test_snapshot_update_not_found(self):
        self.mock_object(volume.api.API, "get_snapshot", fake_snapshot_get)
        updates = {
            "name": "Updated Test Name",
        }
        body = {"snapshot": updates}
        req = fakes.HTTPRequest.blank('/v3/snapshots/not-the-uuid')
        self.assertRaises(exception.SnapshotNotFound, self.controller.update,
                          req, 'not-the-uuid', body=body)

    @mock.patch.object(volume.api.API, "update_snapshot",
                       side_effect=v2_fakes.fake_snapshot_update)
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    @mock.patch('cinder.db.volume_get')
    @mock.patch('cinder.objects.Snapshot.get_by_id')
    def test_snapshot_update_with_leading_trailing_spaces(
            self, snapshot_get_by_id, volume_get,
            snapshot_metadata_get, update_snapshot):
        snapshot = {
            'id': UUID,
            'volume_id': fake.VOLUME_ID,
            'status': fields.SnapshotStatus.AVAILABLE,
            'created_at': "2018-01-14 00:00:00",
            'volume_size': 100,
            'display_name': 'Default name',
            'display_description': 'Default description',
            'expected_attrs': ['metadata'],
        }
        snapshot_obj = fake_snapshot.fake_snapshot_obj(self.ctx, **snapshot)
        fake_volume_obj = fake_volume.fake_volume_obj(self.ctx)
        snapshot_get_by_id.return_value = snapshot_obj
        volume_get.return_value = fake_volume_obj

        updates = {
            "name": "     test     ",
            "description": "     test     "
        }
        body = {"snapshot": updates}
        req = fakes.HTTPRequest.blank('/v3/snapshots/%s' % UUID)
        req.environ['cinder.context'] = self.ctx
        res_dict = self.controller.update(req, UUID, body=body)
        expected = {
            'snapshot': {
                'id': UUID,
                'volume_id': fake.VOLUME_ID,
                'status': fields.SnapshotStatus.AVAILABLE,
                'size': 100,
                'created_at': datetime.datetime(2018, 1, 14, 0, 0, 0,
                                                tzinfo=pytz.utc),
                'updated_at': None,
                'name': u'test',
                'description': u'test',
                'metadata': {},
            }
        }
        self.assertEqual(expected, res_dict)
        self.assertEqual(2, len(self.notifier.notifications))

    @mock.patch.object(volume.api.API, "delete_snapshot",
                       side_effect=v2_fakes.fake_snapshot_update)
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    @mock.patch('cinder.objects.Volume.get_by_id')
    @mock.patch('cinder.objects.Snapshot.get_by_id')
    def test_snapshot_delete(self, snapshot_get_by_id, volume_get_by_id,
                             snapshot_metadata_get, delete_snapshot):
        snapshot = {
            'id': UUID,
            'volume_id': fake.VOLUME_ID,
            'status': fields.SnapshotStatus.AVAILABLE,
            'volume_size': 100,
            'display_name': 'Default name',
            'display_description': 'Default description',
            'expected_attrs': ['metadata'],
        }
        snapshot_obj = fake_snapshot.fake_snapshot_obj(self.ctx, **snapshot)
        fake_volume_obj = fake_volume.fake_volume_obj(self.ctx)
        snapshot_get_by_id.return_value = snapshot_obj
        volume_get_by_id.return_value = fake_volume_obj

        snapshot_id = UUID
        req = fakes.HTTPRequest.blank('/v3/snapshots/%s' % snapshot_id)
        req.environ['cinder.context'] = self.ctx
        resp = self.controller.delete(req, snapshot_id)
        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)

    def test_snapshot_delete_invalid_id(self):
        self.mock_object(volume.api.API, "delete_snapshot",
                         fake_snapshot_delete)
        snapshot_id = INVALID_UUID
        req = fakes.HTTPRequest.blank('/v3/snapshots/%s' % snapshot_id)
        self.assertRaises(exception.SnapshotNotFound, self.controller.delete,
                          req, snapshot_id)

    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    @mock.patch('cinder.objects.Volume.get_by_id')
    @mock.patch('cinder.objects.Snapshot.get_by_id')
    def test_snapshot_show(self, snapshot_get_by_id, volume_get_by_id,
                           snapshot_metadata_get):
        snapshot = {
            'id': UUID,
            'volume_id': fake.VOLUME_ID,
            'status': fields.SnapshotStatus.AVAILABLE,
            'volume_size': 100,
            'display_name': 'Default name',
            'display_description': 'Default description',
            'expected_attrs': ['metadata'],
        }
        snapshot_obj = fake_snapshot.fake_snapshot_obj(self.ctx, **snapshot)
        fake_volume_obj = fake_volume.fake_volume_obj(self.ctx)
        snapshot_get_by_id.return_value = snapshot_obj
        volume_get_by_id.return_value = fake_volume_obj
        req = fakes.HTTPRequest.blank('/v3/snapshots/%s' % UUID)
        req.environ['cinder.context'] = self.ctx
        resp_dict = self.controller.show(req, UUID)

        self.assertIn('snapshot', resp_dict)
        self.assertEqual(UUID, resp_dict['snapshot']['id'])
        self.assertIn('updated_at', resp_dict['snapshot'])

    def test_snapshot_show_invalid_id(self):
        snapshot_id = INVALID_UUID
        req = fakes.HTTPRequest.blank('/v3/snapshots/%s' % snapshot_id)
        self.assertRaises(exception.SnapshotNotFound,
                          self.controller.show, req, snapshot_id)

    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    @mock.patch('cinder.objects.Volume.get_by_id')
    @mock.patch('cinder.objects.Snapshot.get_by_id')
    @mock.patch('cinder.volume.api.API.get_all_snapshots')
    def test_snapshot_detail(self, get_all_snapshots, snapshot_get_by_id,
                             volume_get_by_id, snapshot_metadata_get):
        snapshot = {
            'id': UUID,
            'volume_id': fake.VOLUME_ID,
            'status': fields.SnapshotStatus.AVAILABLE,
            'volume_size': 100,
            'display_name': 'Default name',
            'display_description': 'Default description',
            'expected_attrs': ['metadata']
        }
        ctx = context.RequestContext(fake.PROJECT_ID, fake.USER_ID, True)
        snapshot_obj = fake_snapshot.fake_snapshot_obj(ctx, **snapshot)
        fake_volume_obj = fake_volume.fake_volume_obj(ctx)
        snapshot_get_by_id.return_value = snapshot_obj
        volume_get_by_id.return_value = fake_volume_obj
        snapshots = objects.SnapshotList(objects=[snapshot_obj])
        get_all_snapshots.return_value = snapshots

        req = fakes.HTTPRequest.blank('/v3/snapshots/detail')
        resp_dict = self.controller.detail(req)

        self.assertIn('snapshots', resp_dict)
        resp_snapshots = resp_dict['snapshots']
        self.assertEqual(1, len(resp_snapshots))
        self.assertIn('updated_at', resp_snapshots[0])

        resp_snapshot = resp_snapshots.pop()
        self.assertEqual(UUID, resp_snapshot['id'])

    @mock.patch.object(db, 'snapshot_get_all_by_project',
                       v2_fakes.fake_snapshot_get_all_by_project)
    @mock.patch.object(db, 'snapshot_get_all',
                       v2_fakes.fake_snapshot_get_all)
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    def test_admin_list_snapshots_limited_to_project(self,
                                                     snapshot_metadata_get):
        req = fakes.HTTPRequest.blank('/v3/%s/snapshots' % fake.PROJECT_ID,
                                      use_admin_context=True)
        res = self.controller.index(req)

        self.assertIn('snapshots', res)
        self.assertEqual(1, len(res['snapshots']))

    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    def test_list_snapshots_with_limit_and_offset(self,
                                                  snapshot_metadata_get):
        def list_snapshots_with_limit_and_offset(snaps, is_admin):
            req = fakes.HTTPRequest.blank('/v3/%s/snapshots?limit=1'
                                          '&offset=1' % fake.PROJECT_ID,
                                          use_admin_context=is_admin)
            res = self.controller.index(req)

            self.assertIn('snapshots', res)
            self.assertEqual(1, len(res['snapshots']))
            self.assertEqual(snaps[1].id, res['snapshots'][0]['id'])
            self.assertIn('updated_at', res['snapshots'][0])

            # Test that we get an empty list with an offset greater than the
            # number of items
            req = fakes.HTTPRequest.blank('/v3/snapshots?limit=1&offset=3')
            self.assertEqual({'snapshots': []}, self.controller.index(req))

        volume, snaps = self._create_db_snapshots(3)
        # admin case
        list_snapshots_with_limit_and_offset(snaps, is_admin=True)
        # non-admin case
        list_snapshots_with_limit_and_offset(snaps, is_admin=False)

    @mock.patch.object(db, 'snapshot_get_all_by_project')
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    def test_list_snpashots_with_wrong_limit_and_offset(self,
                                                        mock_metadata_get,
                                                        mock_snapshot_get_all):
        """Test list with negative and non numeric limit and offset."""
        mock_snapshot_get_all.return_value = []

        # Negative limit
        req = fakes.HTTPRequest.blank('/v3/snapshots?limit=-1&offset=1')
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.index,
                          req)

        # Non numeric limit
        req = fakes.HTTPRequest.blank('/v3/snapshots?limit=a&offset=1')
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.index,
                          req)

        # Negative offset
        req = fakes.HTTPRequest.blank('/v3/snapshots?limit=1&offset=-1')
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.index,
                          req)

        # Non numeric offset
        req = fakes.HTTPRequest.blank('/v3/snapshots?limit=1&offset=a')
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.index,
                          req)

        # Test that we get an exception HTTPBadRequest(400) with an offset
        # greater than the maximum offset value.
        url = '/v3/snapshots?limit=1&offset=323245324356534235'
        req = fakes.HTTPRequest.blank(url)
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.index, req)

    def _assert_list_next(self, expected_query=None, project=fake.PROJECT_ID,
                          **kwargs):
        """Check a page of snapshots list."""
        # Since we are accessing v2 api directly we don't need to specify
        # v2 in the request path, if we did, we'd get /v3/v2 links back
        request_path = '/v3/%s/snapshots' % project
        expected_path = request_path

        # Construct the query if there are kwargs
        if kwargs:
            request_str = request_path + '?' + urllib.urlencode(kwargs)
        else:
            request_str = request_path

        # Make the request
        req = fakes.HTTPRequest.blank(request_str)
        res = self.controller.index(req)

        # We only expect to have a next link if there is an actual expected
        # query.
        if expected_query:
            # We must have the links
            self.assertIn('snapshots_links', res)
            links = res['snapshots_links']

            # Must be a list of links, even if we only get 1 back
            self.assertIsInstance(links, list)
            next_link = links[0]

            # rel entry must be next
            self.assertIn('rel', next_link)
            self.assertIn('next', next_link['rel'])

            # href entry must have the right path
            self.assertIn('href', next_link)
            href_parts = urllib.urlparse(next_link['href'])
            self.assertEqual(expected_path, href_parts.path)

            # And the query from the next link must match what we were
            # expecting
            params = urllib.parse_qs(href_parts.query)
            self.assertDictEqual(expected_query, params)

        # Make sure we don't have links if we were not expecting them
        else:
            self.assertNotIn('snapshots_links', res)

    def _create_db_snapshots(self, num_snaps):
        volume = utils.create_volume(self.ctx, volume_type_id=None)
        snaps = [utils.create_snapshot(self.ctx,
                                       volume.id,
                                       display_name='snap' + str(i))
                 for i in range(num_snaps)]

        self.addCleanup(db.volume_destroy, self.ctx, volume.id)
        for snap in snaps:
            self.addCleanup(db.snapshot_destroy, self.ctx, snap.id)

        snaps.reverse()
        return volume, snaps

    def test_list_snapshots_next_link_default_limit(self):
        """Test that snapshot list pagination is limited by osapi_max_limit."""
        volume, snaps = self._create_db_snapshots(3)

        # NOTE(geguileo): Since cinder.api.common.limited has already been
        # imported his argument max_limit already has a default value of 1000
        # so it doesn't matter that we change it to 2.  That's why we need to
        # mock it and send it current value.  We still need to set the default
        # value because other sections of the code use it, for example
        # _get_collection_links
        CONF.set_default('osapi_max_limit', 2)

        def get_pagination_params(params, max_limit=CONF.osapi_max_limit,
                                  original_call=common.get_pagination_params):
            return original_call(params, max_limit)

        def _get_limit_param(params, max_limit=CONF.osapi_max_limit,
                             original_call=common._get_limit_param):
            return original_call(params, max_limit)

        with mock.patch.object(common, 'get_pagination_params',
                               get_pagination_params), \
                mock.patch.object(common, '_get_limit_param',
                                  _get_limit_param):
            # The link from the first page should link to the second
            self._assert_list_next({'marker': [snaps[1].id]})

            # Second page should have no next link
            self._assert_list_next(marker=snaps[1].id)

    def test_list_snapshots_next_link_with_limit(self):
        """Test snapshot list pagination with specific limit."""
        volume, snaps = self._create_db_snapshots(2)

        # The link from the first page should link to the second
        self._assert_list_next({'limit': ['1'], 'marker': [snaps[0].id]},
                               limit=1)

        # Even though there are no more elements, we should get a next element
        # per specification.
        expected = {'limit': ['1'], 'marker': [snaps[1].id]}
        self._assert_list_next(expected, limit=1, marker=snaps[0].id)

        # When we go beyond the number of elements there should be no more
        # next links
        self._assert_list_next(limit=1, marker=snaps[1].id)

    @mock.patch.object(db, 'snapshot_get_all_by_project',
                       v2_fakes.fake_snapshot_get_all_by_project)
    @mock.patch.object(db, 'snapshot_get_all',
                       v2_fakes.fake_snapshot_get_all)
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    def test_admin_list_snapshots_all_tenants(self, snapshot_metadata_get):
        req = fakes.HTTPRequest.blank('/v3/%s/snapshots?all_tenants=1' %
                                      fake.PROJECT_ID,
                                      use_admin_context=True)
        res = self.controller.index(req)
        self.assertIn('snapshots', res)
        self.assertEqual(3, len(res['snapshots']))

    @mock.patch.object(db, 'snapshot_get_all')
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    def test_admin_list_snapshots_by_tenant_id(self, snapshot_metadata_get,
                                               snapshot_get_all):
        def get_all(context, filters=None, marker=None, limit=None,
                    sort_keys=None, sort_dirs=None, offset=None):
            if 'project_id' in filters and 'tenant1' in filters['project_id']:
                return [v2_fakes.fake_snapshot(fake.VOLUME_ID,
                                               tenant_id='tenant1')]
            else:
                return []

        snapshot_get_all.side_effect = get_all

        req = fakes.HTTPRequest.blank('/v3/%s/snapshots?all_tenants=1'
                                      '&project_id=tenant1' % fake.PROJECT_ID,
                                      use_admin_context=True)
        res = self.controller.index(req)
        self.assertIn('snapshots', res)
        self.assertEqual(1, len(res['snapshots']))

    @mock.patch.object(db, 'snapshot_get_all_by_project',
                       v2_fakes.fake_snapshot_get_all_by_project)
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    def test_all_tenants_non_admin_gets_all_tenants(self,
                                                    snapshot_metadata_get):
        req = fakes.HTTPRequest.blank('/v3/%s/snapshots?all_tenants=1' %
                                      fake.PROJECT_ID)
        res = self.controller.index(req)
        self.assertIn('snapshots', res)
        self.assertEqual(1, len(res['snapshots']))

    @mock.patch.object(db, 'snapshot_get_all_by_project',
                       v2_fakes.fake_snapshot_get_all_by_project)
    @mock.patch.object(db, 'snapshot_get_all',
                       v2_fakes.fake_snapshot_get_all)
    @mock.patch('cinder.db.snapshot_metadata_get', return_value=dict())
    def test_non_admin_get_by_project(self, snapshot_metadata_get):
        req = fakes.HTTPRequest.blank('/v3/%s/snapshots' % fake.PROJECT_ID)
        res = self.controller.index(req)
        self.assertIn('snapshots', res)
        self.assertEqual(1, len(res['snapshots']))

    def _create_snapshot_bad_body(self, body):
        req = fakes.HTTPRequest.blank('/v3/%s/snapshots' % fake.PROJECT_ID)
        req.method = 'POST'

        self.assertRaises(exception.ValidationError,
                          self.controller.create, req, body=body)

    def test_create_no_body(self):
        self._create_snapshot_bad_body(body=None)

    def test_create_missing_snapshot(self):
        body = {'foo': {'a': 'b'}}
        self._create_snapshot_bad_body(body=body)

    def test_create_malformed_entity(self):
        body = {'snapshot': 'string'}
        self._create_snapshot_bad_body(body=body)
